import os
import torch
from pathlib import Path
import torch.utils.data as Data
from typing import Optional

from vame.schemas.states import EvaluateModelFunctionSchema, save_state
from vame.model.rnn_vae import RNN_VAE
from vame.model.dataloader import SEQUENCE_DATASET
from vame.logging.logger import VameLogger
from vame.visualization.model import (
    plot_reconstruction,
    plot_loss,
)


logger_config = VameLogger(__name__)
logger = logger_config.logger

use_gpu = torch.cuda.is_available()
if use_gpu:
    pass
else:
    torch.device("cpu")


def eval_temporal(
    cfg: dict,
    use_gpu: bool,
    model_name: str,
    fixed: bool,
    snapshot: Optional[str] = None,
    suffix: Optional[str] = None,
) -> None:
    """
    Evaluate the temporal aspects of the trained model.

    Parameters
    ----------
    cfg : dict
        Configuration dictionary.
    use_gpu : bool
        Flag indicating whether to use GPU for evaluation.
    model_name : str
        Name of the model.
    fixed : bool
        Flag indicating whether the data is fixed or not.
    snapshot : str, optional
        Path to the model snapshot. Defaults to None.
    suffix : str, optional
        Suffix for the saved plot filename. Defaults to None.

    Returns
    -------
    None
    """
    SEED = 19
    ZDIMS = cfg["zdims"]
    FUTURE_DECODER = cfg["prediction_decoder"]
    TEMPORAL_WINDOW = cfg["time_window"] * 2
    FUTURE_STEPS = cfg["prediction_steps"]
    NUM_FEATURES = cfg["num_features"]
    if not fixed:
        NUM_FEATURES = NUM_FEATURES - 3
    TEST_BATCH_SIZE = 64
    hidden_size_layer_1 = cfg["hidden_size_layer_1"]
    hidden_size_layer_2 = cfg["hidden_size_layer_2"]
    hidden_size_rec = cfg["hidden_size_rec"]
    hidden_size_pred = cfg["hidden_size_pred"]
    dropout_encoder = cfg["dropout_encoder"]
    dropout_rec = cfg["dropout_rec"]
    dropout_pred = cfg["dropout_pred"]
    softplus = cfg["softplus"]

    filepath = os.path.join(cfg["project_path"], "model")

    seq_len_half = int(TEMPORAL_WINDOW / 2)
    if use_gpu:
        torch.cuda.manual_seed(SEED)
        model = RNN_VAE(
            TEMPORAL_WINDOW,
            ZDIMS,
            NUM_FEATURES,
            FUTURE_DECODER,
            FUTURE_STEPS,
            hidden_size_layer_1,
            hidden_size_layer_2,
            hidden_size_rec,
            hidden_size_pred,
            dropout_encoder,
            dropout_rec,
            dropout_pred,
            softplus,
        ).cuda()
        model.load_state_dict(
            torch.load(
                os.path.join(
                    cfg["project_path"],
                    "model",
                    "best_model",
                    model_name + "_" + cfg["project_name"] + ".pkl",
                )
            )
        )
    else:
        model = RNN_VAE(
            TEMPORAL_WINDOW,
            ZDIMS,
            NUM_FEATURES,
            FUTURE_DECODER,
            FUTURE_STEPS,
            hidden_size_layer_1,
            hidden_size_layer_2,
            hidden_size_rec,
            hidden_size_pred,
            dropout_encoder,
            dropout_rec,
            dropout_pred,
            softplus,
        ).to()
        if not snapshot:
            model.load_state_dict(
                torch.load(
                    os.path.join(
                        cfg["project_path"],
                        "model",
                        "best_model",
                        model_name + "_" + cfg["project_name"] + ".pkl",
                    ),
                    map_location=torch.device("cpu"),
                )
            )
        elif snapshot:
            model.load_state_dict(torch.load(snapshot), map_location=torch.device("cpu"))
    model.eval()  # toggle evaluation mode

    testset = SEQUENCE_DATASET(
        os.path.join(cfg["project_path"], "data", "train", ""),
        data="test_seq.npy",
        train=False,
        temporal_window=TEMPORAL_WINDOW,
        logger_config=logger_config,
    )
    test_loader = Data.DataLoader(testset, batch_size=TEST_BATCH_SIZE, shuffle=True, drop_last=True)

    if not snapshot:
        plot_reconstruction(
            filepath,
            test_loader,
            seq_len_half,
            model,
            model_name,
            FUTURE_DECODER,
            FUTURE_STEPS,
        )  # , suffix=suffix
    elif snapshot:
        plot_reconstruction(
            filepath,
            test_loader,
            seq_len_half,
            model,
            model_name,
            FUTURE_DECODER,
            FUTURE_STEPS,
            suffix=suffix,
        )  # ,
    # if use_gpu:
    #     plot_loss(
    #         cfg=cfg,
    #         filepath=filepath,
    #         model_name=model_name,
    #         show_figure=False,
    #     )
    # else:
    #     plot_loss(
    #         cfg=cfg,
    #         filepath=filepath,
    #         model_name=model_name,
    #         show_figure=False,
    #     )
    #     # pass #note, loading of losses needs to be adapted for CPU use #TODO


@save_state(model=EvaluateModelFunctionSchema)
def evaluate_model(
    config: dict,
    use_snapshots: bool = False,
    save_logs: bool = False,
) -> None:
    """
    Evaluate the trained model.
    Fills in the values in the "evaluate_model" key of the states.json file.
    Saves the evaluation results to:
    - project_name/
        - model/
            - evaluate/

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    use_snapshots : bool, optional
        Whether to plot for all snapshots or only the best model. Defaults to False.
    save_logs : bool, optional
        Flag indicating whether to save logs. Defaults to False.

    Returns
    -------
    None
    """
    project_path = Path(config["project_path"]).resolve()
    try:
        if save_logs:
            log_path = project_path / "logs" / "evaluate_model.log"
            logger_config.add_file_handler(str(log_path))

        model_name = config["model_name"]
        fixed = config["egocentric_data"]

        use_gpu = torch.cuda.is_available()
        if use_gpu:
            logger.info("Using CUDA")
            logger.info("GPU active: {}".format(torch.cuda.is_available()))
            logger.info("GPU used: {}".format(torch.cuda.get_device_name(0)))
        else:
            torch.device("cpu")
            logger.info("CUDA is not working, or a GPU is not found; using CPU!")

        logger.info(f"Evaluation of model: {model_name}")
        if not use_snapshots:
            eval_temporal(
                cfg=config,
                use_gpu=use_gpu,
                model_name=model_name,
                fixed=fixed,
            )
        elif use_snapshots:
            snapshots = os.listdir(os.path.join(str(project_path), "model", "best_model", "snapshots"))
            for snap in snapshots:
                fullpath = os.path.join(str(project_path), "model", "best_model", "snapshots", snap)
                epoch = snap.split("_")[-1]
                eval_temporal(
                    cfg=config,
                    use_gpu=use_gpu,
                    model_name=model_name,
                    fixed=fixed,
                    snapshot=fullpath,
                    suffix="snapshot" + str(epoch),
                )

        logger.info(
            "You can find the results of the evaluation in '/Your-VAME-Project/model/evaluate/' \n"
            "OPTIONS:\n"
            "- vame.segment_session() to identify behavioral motifs.\n"
            "- re-run the model for further fine tuning. Check again with vame.evaluate_model()"
        )
    except Exception as e:
        logger.exception(f"An error occurred during model evaluation: {e}")
    finally:
        logger_config.remove_file_handler()



import xarray as xr


def format_xarray_for_rnn(
    ds: xr.Dataset,
    read_from_variable: str = "position_processed",
):
    """
    Formats the xarray dataset for use VAME's RNN model:
    - The x and y coordinates of the centered_reference_keypoint are excluded.
    - The x coordinate of the orientation_reference_keypoint is excluded.
    - The remaining data is flattened and transposed.

    Parameters
    ----------
    ds : xr.Dataset
        The xarray dataset to format.
    read_from_variable : str, default="position_processed"
        The variable to read from the dataset.

    Returns
    -------
    np.ndarray
        The formatted array in the shape (n_features, n_samples).
        Where n_features = 2 * n_keypoints * n_spaces - 3.
    """
    data = ds[read_from_variable]
    centered_reference_keypoint = ds.attrs["centered_reference_keypoint"]
    orientation_reference_keypoint = ds.attrs["orientation_reference_keypoint"]

    # Get the coordinates
    individuals = data.coords["individuals"].values
    keypoints = data.coords["keypoints"].values
    spaces = data.coords["space"].values

    # Create a flattened array and infer column indices
    flattened_array = data.values.reshape(data.shape[0], -1)
    columns = [f"{ind}_{kp}_{sp}" for ind in individuals for kp in keypoints for sp in spaces]

    # Identify columns to exclude
    excluded_columns = []
    for ind in individuals:
        # Exclude both x and y for centered_reference_keypoint
        excluded_columns.append(f"{ind}_{centered_reference_keypoint}_x")
        excluded_columns.append(f"{ind}_{centered_reference_keypoint}_y")
        # Exclude only x for orientation_reference_keypoint
        excluded_columns.append(f"{ind}_{orientation_reference_keypoint}_x")

    # Filter out the excluded columns
    included_indices = [i for i, col in enumerate(columns) if col not in excluded_columns]
    filtered_array = flattened_array[:, included_indices]

    return filtered_array.T



import os
import numpy as np
from pathlib import Path
from typing import List

from vame.logging.logger import VameLogger
from vame.schemas.states import CreateTrainsetFunctionSchema, save_state
from vame.io.load_poses import read_pose_estimation_file
from vame.preprocessing.to_model import format_xarray_for_rnn


logger_config = VameLogger(__name__)
logger = logger_config.logger


def traindata_aligned(
    config: dict,
    sessions: List[str] | None = None,
    test_fraction: float | None = None,
    read_from_variable: str = "position_processed",
) -> None:
    """
    Create training dataset for aligned data.
    Save numpy arrays with the test/train info to the project folder.

    Parameters
    ----------
    config : dict
        Configuration parameters dictionary.
    sessions : List[str], optional
        List of session names. If None, all sessions will be used. Defaults to None.
    test_fraction : float, optional
        Fraction of data to use as test data. Defaults to 0.1.

    Returns
    -------
    None
    """
    project_path = config["project_path"]
    if sessions is None:
        sessions = config["session_names"]
    if test_fraction is None:
        test_fraction = config["test_fraction"]

    all_data_list = []
    for session in sessions:
        # Read session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)

        # Format the data for the RNN model
        filtered_array = format_xarray_for_rnn(
            ds=ds,
            read_from_variable=read_from_variable,
        )

        all_data_list.append(filtered_array)

    all_data_array = np.concatenate(all_data_list, axis=1)
    test_size = int(all_data_array.shape[1] * test_fraction)
    data_test = all_data_array[:, :test_size]
    data_train = all_data_array[:, test_size:]

    # Save numpy arrays the the test/train info:
    train_data_path = Path(project_path) / "data" / "train" / "train_seq.npy"
    np.save(str(train_data_path), data_train)

    test_data_path = Path(project_path) / "data" / "train" / "test_seq.npy"
    np.save(str(test_data_path), data_test)

    logger.info(f"Lenght of train data: {data_train.shape[1]}")
    logger.info(f"Lenght of test data: {data_test.shape[1]}")


# def traindata_fixed(
#     cfg: dict,
#     sessions: List[str],
#     testfraction: float,
#     num_features: int,
#     savgol_filter: bool,
#     check_parameter: bool,
#     pose_ref_index: Optional[List[int]],
# ) -> None:
#     """
#     Create training dataset for fixed data.

#     Parameters
#     ---------
#     cfg : dict
#         Configuration parameters.
#     sessions : List[str]
#         List of sessions.
#     testfraction : float
#         Fraction of data to use as test data.
#     num_features : int
#         Number of features.
#     savgol_filter : bool
#         Flag indicating whether to apply Savitzky-Golay filter.
#     check_parameter : bool
#         If True, the function will plot the z-scored data and the filtered data.
#     pose_ref_index : Optional[List[int]]
#         List of reference coordinate indices for alignment.

#     Returns
#         None
#             Save numpy arrays with the test/train info to the project folder.
#     """
#     X_train = []
#     pos = []
#     pos_temp = 0
#     pos.append(0)

#     if check_parameter:
#         X_true = []
#         sessions = [sessions[0]]

#     for session in sessions:
#         logger.info("z-scoring of file %s" % session)
#         path_to_file = os.path.join(
#             cfg["project_path"],
#             "data",
#             "processed",
#             session,
#             session + "-PE-seq.npy",
#         )
#         data = np.load(path_to_file)

#         X_mean = np.mean(data, axis=None)
#         X_std = np.std(data, axis=None)
#         X_z = (data.T - X_mean) / X_std

#         if check_parameter:
#             X_z_copy = X_z.copy()
#             X_true.append(X_z_copy)

#         if cfg["robust"]:
#             iqr_val = iqr(X_z)
#             logger.info("IQR value: %.2f, IQR cutoff: %.2f" % (iqr_val, cfg["iqr_factor"] * iqr_val))
#             for i in range(X_z.shape[0]):
#                 for marker in range(X_z.shape[1]):
#                     if X_z[i, marker] > cfg["iqr_factor"] * iqr_val:
#                         X_z[i, marker] = np.nan

#                     elif X_z[i, marker] < -cfg["iqr_factor"] * iqr_val:
#                         X_z[i, marker] = np.nan

#                 X_z[i, :] = interpol_all_nans(X_z[i, :])

#         X_len = len(data.T)
#         pos_temp += X_len
#         pos.append(pos_temp)
#         X_train.append(X_z)

#     X = np.concatenate(X_train, axis=0).T

#     if savgol_filter:
#         X_med = scipy.signal.savgol_filter(X, cfg["savgol_length"], cfg["savgol_order"])
#     else:
#         X_med = X

#     num_frames = len(X_med.T)
#     test = int(num_frames * testfraction)

#     z_test = X_med[:, :test]
#     z_train = X_med[:, test:]

#     if check_parameter:
#         plot_check_parameter(
#             cfg,
#             iqr_val,
#             num_frames,
#             X_true,
#             X_med,
#         )

#     else:
#         if pose_ref_index is None:
#             raise ValueError("Please provide a pose reference index for training on fixed data. E.g. [0,5]")
#         # save numpy arrays the the test/train info:
#         np.save(
#             os.path.join(
#                 cfg["project_path"],
#                 "data",
#                 "train",
#                 "train_seq.npy",
#             ),
#             z_train,
#         )
#         np.save(
#             os.path.join(
#                 cfg["project_path"],
#                 "data",
#                 "train",
#                 "test_seq.npy",
#             ),
#             z_test,
#         )

#         y_shifted_indices = np.arange(0, num_features, 2)
#         x_shifted_indices = np.arange(1, num_features, 2)
#         belly_Y_ind = pose_ref_index[0] * 2
#         belly_X_ind = (pose_ref_index[0] * 2) + 1

#         for i, session in enumerate(sessions):
#             # Shifting section added 2/29/2024 PN
#             X_med_shifted_file = X_med[:, pos[i] : pos[i + 1]]
#             belly_Y_shift = X_med[belly_Y_ind, pos[i] : pos[i + 1]]
#             belly_X_shift = X_med[belly_X_ind, pos[i] : pos[i + 1]]

#             X_med_shifted_file[y_shifted_indices, :] -= belly_Y_shift
#             X_med_shifted_file[x_shifted_indices, :] -= belly_X_shift

#             np.save(
#                 os.path.join(
#                     cfg["project_path"],
#                     "data",
#                     "processed",
#                     session,
#                     session + "-PE-seq-clean.npy",
#                 ),
#                 X_med_shifted_file,
#             )  # saving new shifted file

#         logger.info("Lenght of train data: %d" % len(z_train.T))
#         logger.info("Lenght of test data: %d" % len(z_test.T))


@save_state(model=CreateTrainsetFunctionSchema)
def create_trainset(
    config: dict,
    save_logs: bool = False,
) -> None:
    """
    Creates a training and test datasets for the VAME model.
    Fills in the values in the "create_trainset" key of the states.json file.
    Creates the training dataset for VAME at:
    - project_name/
        - data/
            - session00/
                - session00-PE-seq-clean.npy
            - session01/
                - session01-PE-seq-clean.npy
            - train/
                - test_seq.npy
                - train_seq.npy

    The produced -clean.npy files contain the aligned time series data in the
    shape of (num_dlc_features - 2, num_video_frames).

    The produced test_seq.npy contains the combined data in the shape of (num_dlc_features - 2, num_video_frames * test_fraction).

    The produced train_seq.npy contains the combined data in the shape of (num_dlc_features - 2, num_video_frames * (1 - test_fraction)).

    Parameters
    ----------
    config : dict
        Configuration parameters dictionary.
    save_logs : bool, optional
        If True, the function will save logs to the project folder. Defaults to False.

    Returns
    -------
    None
    """
    try:
        fixed = config["egocentric_data"]

        if save_logs:
            log_path = Path(config["project_path"]) / "logs" / "create_trainset.log"
            logger_config.add_file_handler(str(log_path))

        if not os.path.exists(os.path.join(config["project_path"], "data", "train", "")):
            os.mkdir(os.path.join(config["project_path"], "data", "train", ""))

        sessions = []
        if config["all_data"] == "No":
            for session in config["session_names"]:
                use_session = input("Do you want to train on " + session + "? yes/no: ")
                if use_session == "yes":
                    sessions.append(session)
                if use_session == "no":
                    continue
        else:
            sessions = config["session_names"]

        logger.info("Creating training dataset...")

        if not fixed:
            logger.info("Creating trainset from the vame.egocentrical_alignment() output ")
            traindata_aligned(
                config=config,
                sessions=sessions,
            )
        else:
            raise NotImplementedError("Fixed data training is not implemented yet")
            # logger.info("Creating trainset from the vame.pose_to_numpy() output ")
            # traindata_fixed(
            #     cfg,
            #     sessions,
            #     cfg["test_fraction"],
            #     cfg["num_features"],
            #     cfg["savgol_filter"],
            #     check_parameter,
            #     pose_ref_index,
            # )

        logger.info("A training and test set has been created. Next step: vame.train_model()")

    except Exception as e:
        logger.exception(str(e))
        raise e
    finally:
        logger_config.remove_file_handler()



import numpy as np
from pathlib import Path

from vame.logging.logger import VameLogger
from vame.io.load_poses import read_pose_estimation_file


logger_config = VameLogger(__name__)
logger = logger_config.logger


def egocentrically_align_and_center(
    config: dict,
    centered_reference_keypoint: str = "snout",
    orientation_reference_keypoint: str = "tailbase",
    read_from_variable: str = "position_processed",
    save_to_variable: str = "position_egocentric_aligned",
) -> None:
    """
    Aligns the time series by first centralizing all positions around the first keypoint
    and then applying rotation to align with the line connecting the two keypoints.

    Parameters
    ----------
    config : dict
        Configuration dictionary
    centered_reference_keypoint : str
        Name of the keypoint to use as centered reference.
    orientation_reference_keypoint : str
        Name of the keypoint to use as orientation reference.

    Returns
    -------
    None
    """
    logger.info(
        f"Egocentric alignment with references: {centered_reference_keypoint} and {orientation_reference_keypoint}"
    )
    project_path = config["project_path"]
    sessions = config["session_names"]

    for i, session in enumerate(sessions):
        logger.info(f"Session: {session}")
        # Read session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)

        # Extract keypoint indices
        keypoints = ds.coords["keypoints"].values
        if centered_reference_keypoint not in keypoints:
            raise ValueError(
                f"Centered reference keypoint {centered_reference_keypoint} not found in dataset.",
                f"Available keypoints: {keypoints}",
            )
        if orientation_reference_keypoint not in keypoints:
            raise ValueError(
                f"Orientation reference keypoint {orientation_reference_keypoint} not found in dataset.",
                f"Available keypoints: {keypoints}",
            )
        idx1 = np.where(keypoints == centered_reference_keypoint)[0][0]
        idx2 = np.where(keypoints == orientation_reference_keypoint)[0][0]

        # Extract processed positions values, with shape: (time, individuals, keypoints, space)
        position_processed = np.copy(ds[read_from_variable].values)
        position_aligned = np.empty_like(position_processed)

        # Loop over individuals
        for individual in range(position_processed.shape[1]):
            # Shape: (time, keypoints, space)
            individual_positions = position_processed[:, individual, :, :]
            centralized_positions = np.empty_like(individual_positions)

            # Centralize all positions around the first keypoint
            for kp in range(individual_positions.shape[1]):
                for space in range(individual_positions.shape[2]):
                    centralized_positions[:, kp, space] = (
                        individual_positions[:, kp, space] - individual_positions[:, idx1, space]
                    )

            # Calculate vectors between keypoints
            vector = centralized_positions[:, idx2, :]  # Vector from keypoint1 to keypoint2
            angles = np.arctan2(vector[:, 0], vector[:, 1])  # Angles in radians

            # Apply rotation to align the second keypoint along the y-axis
            for t in range(centralized_positions.shape[0]):
                rotation_matrix = np.array(
                    [
                        [np.cos(angles[t]), -np.sin(angles[t])],
                        [np.sin(angles[t]), np.cos(angles[t])],
                    ]
                )
                frame_positions = centralized_positions[t, :, :]
                rotated_positions = (rotation_matrix @ frame_positions.T).T

                # Check and ensure the y-value of orientation_reference_keypoint is negative
                if rotated_positions[idx2, 1] > 0:
                    rotated_positions[:, :] *= -1  # Flip all coordinates

                position_aligned[t, individual, :, :] = rotated_positions

        # Update the dataset with the cleaned position values
        ds[save_to_variable] = (ds[read_from_variable].dims, position_aligned)
        ds.attrs.update(
            {
                "processed_alignment": True,
                "centered_reference_keypoint": centered_reference_keypoint,
                "orientation_reference_keypoint": orientation_reference_keypoint,
            }
        )

        # Save the aligned dataset to file
        cleaned_file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        ds.to_netcdf(
            path=cleaned_file_path,
            engine="scipy",
        )



import os
import tqdm
import torch
import pickle
import numpy as np
from pathlib import Path
from typing import List, Tuple, Union
from hmmlearn import hmm
from sklearn.cluster import KMeans

from vame.schemas.states import save_state, SegmentSessionFunctionSchema
from vame.logging.logger import VameLogger, TqdmToLogger
from vame.model.rnn_model import RNN_VAE
from vame.io.load_poses import read_pose_estimation_file
from vame.util.cli import get_sessions_from_user_input
from vame.util.model_util import load_model
from vame.preprocessing.to_model import format_xarray_for_rnn


logger_config = VameLogger(__name__)
logger = logger_config.logger


def embedd_latent_vectors(
    cfg: dict,
    sessions: List[str],
    model: RNN_VAE,
    fixed: bool,
    read_from_variable: str = "position_processed",
    tqdm_stream: Union[TqdmToLogger, None] = None,
) -> List[np.ndarray]:
    """
    Embed latent vectors for the given files using the VAME model.

    Parameters
    ----------
    cfg : dict
        Configuration dictionary.
    sessions : List[str]
        List of session names.
    model : RNN_VAE
        VAME model.
    fixed : bool
        Whether the model is fixed.
    tqdm_stream : TqdmToLogger, optional
        TQDM Stream to redirect the tqdm output to logger.

    Returns
    -------
    List[np.ndarray]
        List of latent vectors for each file.
    """
    project_path = cfg["project_path"]
    temp_win = cfg["time_window"]
    num_features = cfg["num_features"]
    if not fixed:
        num_features = num_features - 3

    use_gpu = torch.cuda.is_available()
    if use_gpu:
        pass
    else:
        torch.device("cpu")

    latent_vector_files = []

    for session in sessions:
        logger.info(f"Embedding of latent vector for file {session}")
        # Read session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)
        data = np.copy(ds[read_from_variable].values)

        # Format the data for the RNN model
        data = format_xarray_for_rnn(
            ds=ds,
            read_from_variable=read_from_variable,
        )

        latent_vector_list = []
        with torch.no_grad():
            for i in tqdm.tqdm(range(data.shape[1] - temp_win), file=tqdm_stream):
                # for i in tqdm.tqdm(range(10000)):
                data_sample_np = data[:, i : temp_win + i].T
                data_sample_np = np.reshape(data_sample_np, (1, temp_win, num_features))
                if use_gpu:
                    h_n = model.encoder(torch.from_numpy(data_sample_np).type("torch.FloatTensor").cuda())
                else:
                    h_n = model.encoder(torch.from_numpy(data_sample_np).type("torch.FloatTensor").to())
                mu, _, _ = model.lmbda(h_n)
                latent_vector_list.append(mu.cpu().data.numpy())

        latent_vector = np.concatenate(latent_vector_list, axis=0)
        latent_vector_files.append(latent_vector)

    return latent_vector_files


def get_motif_usage(
    session_labels: np.ndarray,
    n_clusters: int,
) -> np.ndarray:
    """
    Count motif usage from session label array.

    Parameters
    ----------
    session_labels : np.ndarray
        Array of session labels.
    n_clusters : int
        Number of clusters.

    Returns
    -------
    np.ndarray
        Array of motif usage counts.
    """
    motif_usage = np.zeros(n_clusters)
    for i in range(n_clusters):
        motif_count = np.sum(session_labels == i)
        motif_usage[i] = motif_count
    # Include warning if any unused motifs are present
    unused_motifs = np.where(motif_usage == 0)[0]
    if unused_motifs.size > 0:
        logger.info(f"Warning: The following motifs are unused: {unused_motifs}")
    return motif_usage


def same_segmentation(
    cfg: dict,
    sessions: List[str],
    latent_vectors: List[np.ndarray],
    n_clusters: int,
    segmentation_algorithm: str,
) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """
    Apply the same segmentation to all animals.

    Parameters
    ----------
    cfg : dict
        Configuration dictionary.
    sessions : List[str]
        List of session names.
    latent_vectors : List[np.ndarray]
        List of latent vector arrays.
    n_clusters : int
        Number of clusters.
    segmentation_algorithm : str
        Segmentation algorithm.

    Returns
    -------
    Tuple
        Tuple of labels, cluster centers, and motif usages.
    """
    # List of arrays containing each session's motif labels #[SRM, 10/28/24], recommend rename this and similar variables to allsessions_labels
    labels = []
    cluster_centers = []  # List of arrays containing each session's cluster centers
    motif_usages = []  # List of arrays containing each session's motif usages

    latent_vector_cat = np.concatenate(latent_vectors, axis=0)
    if segmentation_algorithm == "kmeans":
        logger.info("Using kmeans as segmentation algorithm!")
        kmeans = KMeans(
            init="k-means++",
            n_clusters=n_clusters,
            random_state=42,
            n_init=20,
        ).fit(latent_vector_cat)
        clust_center = kmeans.cluster_centers_
        # 1D, vector of all labels for the entire cohort
        label = kmeans.predict(latent_vector_cat)

    elif segmentation_algorithm == "hmm":
        if not cfg["hmm_trained"]:
            logger.info("Using a HMM as segmentation algorithm!")
            hmm_model = hmm.GaussianHMM(
                n_components=n_clusters,
                covariance_type="full",
                n_iter=100,
            )
            hmm_model.fit(latent_vector_cat)
            label = hmm_model.predict(latent_vector_cat)
            save_data = os.path.join(cfg["project_path"], "results", "")
            with open(save_data + "hmm_trained.pkl", "wb") as file:
                pickle.dump(hmm_model, file)
        else:
            logger.info("Using a pretrained HMM as segmentation algorithm!")
            save_data = os.path.join(cfg["project_path"], "results", "")
            with open(save_data + "hmm_trained.pkl", "rb") as file:
                hmm_model = pickle.load(file)
            label = hmm_model.predict(latent_vector_cat)

    idx = 0  # start index for each session
    for i, session in enumerate(sessions):
        logger.info(f"Getting motif usage for {session}")
        file_len = latent_vectors[i].shape[0]  # stop index of the session
        labels.append(label[idx : idx + file_len])  # append session's label
        if segmentation_algorithm == "kmeans":
            cluster_centers.append(clust_center)

        # session's motif usage
        motif_usage = get_motif_usage(label[idx : idx + file_len], n_clusters)
        motif_usages.append(motif_usage)
        idx += file_len  # updating the session start index

    return labels, cluster_centers, motif_usages


def individual_segmentation(
    cfg: dict,
    sessions: List[str],
    latent_vectors: List[np.ndarray],
    n_clusters: int,
) -> Tuple:
    """
    Apply individual segmentation to each session.

    Parameters
    ----------
    cfg : dict
        Configuration dictionary.
    sessions : List[str]
        List of session names.
    latent_vectors : List[np.ndarray]
        List of latent vector arrays.
    n_clusters : int
        Number of clusters.

    Returns
    -------
    Tuple
        Tuple of labels, cluster centers, and motif usages.
    """
    random_state = cfg["random_state_kmeans"]
    n_init = cfg["n_init_kmeans"]
    labels = []
    cluster_centers = []
    motif_usages = []
    for i, session in enumerate(sessions):
        logger.info(f"Processing session: {session}")
        kmeans = KMeans(
            init="k-means++",
            n_clusters=n_clusters,
            random_state=random_state,
            n_init=n_init,
        ).fit(latent_vectors[i])
        clust_center = kmeans.cluster_centers_
        label = kmeans.predict(latent_vectors[i])
        motif_usage = get_motif_usage(
            session_labels=label,
            n_clusters=n_clusters,
        )
        motif_usages.append(motif_usage)
        labels.append(label)
        cluster_centers.append(clust_center)
    return labels, cluster_centers, motif_usages


@save_state(model=SegmentSessionFunctionSchema)
def segment_session(
    config: dict,
    save_logs: bool = False,
) -> None:
    """
    Perform pose segmentation using the VAME model.
    Fills in the values in the "segment_session" key of the states.json file.
    Creates files at:
    - project_name/
        - results/
            - hmm_trained.pkl
            - session/
                - model_name/
                    - hmm-n_clusters/
                        - latent_vector_session.npy
                        - motif_usage_session.npy
                        - n_cluster_label_session.npy
                    - kmeans-n_clusters/
                        - latent_vector_session.npy
                        - motif_usage_session.npy
                        - n_cluster_label_session.npy
                        - cluster_center_session.npy

    latent_vector_session.npy contains the projection of the data into the latent space,
    for each frame of the video. Dimmentions: (n_frames, n_latent_features)

    motif_usage_session.npy contains the number of times each motif was used in the video.
    Dimmentions: (n_motifs,)

    n_cluster_label_session.npy contains the label of the cluster assigned to each frame.
    Dimmentions: (n_frames,)

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    save_logs : bool, optional
        Whether to save logs, by default False.

    Returns
    -------
    None
    """
    project_path = Path(config["project_path"]).resolve()
    try:
        tqdm_stream = None
        if save_logs:
            log_path = project_path / "logs" / "pose_segmentation.log"
            logger_config.add_file_handler(str(log_path))
            tqdm_stream = TqdmToLogger(logger)
        model_name = config["model_name"]
        n_clusters = config["n_clusters"]
        fixed = config["egocentric_data"]
        segmentation_algorithms = config["segmentation_algorithms"]
        ind_seg = config["individual_segmentation"]

        logger.info("Pose segmentation for VAME model: %s \n" % model_name)
        logger.info(f"Segmentation algorithms: {segmentation_algorithms}")

        for seg in segmentation_algorithms:
            logger.info(f"Running pose segmentation using {seg} algorithm...")
            for session in config["session_names"]:
                if not os.path.exists(
                    os.path.join(
                        str(project_path),
                        "results",
                        session,
                        model_name,
                        "",
                    )
                ):
                    os.mkdir(
                        os.path.join(
                            str(project_path),
                            "results",
                            session,
                            model_name,
                            "",
                        )
                    )

            # Get sessions
            if config["all_data"] in ["Yes", "yes"]:
                sessions = config["session_names"]
            else:
                sessions = get_sessions_from_user_input(
                    cfg=config,
                    action_message="run segmentation",
                )

            use_gpu = torch.cuda.is_available()
            if use_gpu:
                logger.info("Using CUDA")
                logger.info("GPU active: {}".format(torch.cuda.is_available()))
                logger.info("GPU used: {}".format(torch.cuda.get_device_name(0)))
            else:
                logger.info("CUDA is not working! Attempting to use the CPU...")
                torch.device("cpu")

            if not os.path.exists(
                os.path.join(
                    str(project_path),
                    "results",
                    sessions[0],
                    model_name,
                    seg + "-" + str(n_clusters),
                    "",
                )
            ):
                new = True
                model = load_model(config, model_name, fixed)
                latent_vectors = embedd_latent_vectors(
                    config,
                    sessions,
                    model,
                    fixed,
                    tqdm_stream=tqdm_stream,
                )

                if ind_seg:
                    logger.info(
                        f"Apply individual segmentation of latent vectors for each session, {n_clusters} clusters"
                    )
                    labels, cluster_center, motif_usages = individual_segmentation(
                        cfg=config,
                        sessions=sessions,
                        latent_vectors=latent_vectors,
                        n_clusters=n_clusters,
                    )
                else:
                    logger.info(
                        f"Apply the same segmentation of latent vectors for all sessions, {n_clusters} clusters"
                    )
                    labels, cluster_center, motif_usages = same_segmentation(
                        cfg=config,
                        sessions=sessions,
                        latent_vectors=latent_vectors,
                        n_clusters=n_clusters,
                        segmentation_algorithm=seg,
                    )

            else:
                logger.info(f"\nSegmentation with {n_clusters} k-means clusters already exists for model {model_name}")

                if os.path.exists(
                    os.path.join(
                        str(project_path),
                        "results",
                        sessions[0],
                        model_name,
                        seg + "-" + str(n_clusters),
                        "",
                    )
                ):
                    flag = input(
                        "WARNING: A segmentation for the chosen model and cluster size already exists! \n"
                        "Do you want to continue? A new segmentation will be computed! (yes/no) "
                    )
                else:
                    flag = "yes"

                if flag == "yes":
                    new = True
                    latent_vectors = []
                    for session in sessions:
                        path_to_latent_vector = os.path.join(
                            str(project_path),
                            "results",
                            session,
                            model_name,
                            seg + "-" + str(n_clusters),
                            "",
                        )
                        latent_vector = np.load(
                            os.path.join(
                                path_to_latent_vector,
                                "latent_vector_" + session + ".npy",
                            )
                        )
                        latent_vectors.append(latent_vector)

                    if ind_seg:
                        logger.info(
                            f"Apply individual segmentation of latent vectors for each session, {n_clusters} clusters"
                        )
                        # [SRM, 10/28/24] rename to cluster_centers
                        labels, cluster_center, motif_usages = individual_segmentation(
                            cfg=config,
                            sessions=sessions,
                            latent_vectors=latent_vectors,
                            n_clusters=n_clusters,
                        )
                    else:
                        logger.info(
                            f"Apply the same segmentation of latent vectors for all sessions, {n_clusters} clusters"
                        )
                        # [SRM, 10/28/24] rename to cluster_centers
                        labels, cluster_center, motif_usages = same_segmentation(
                            cfg=config,
                            sessions=sessions,
                            latent_vectors=latent_vectors,
                            n_clusters=n_clusters,
                            segmentation_algorithm=seg,
                        )
                else:
                    logger.info("No new segmentation has been calculated.")
                    new = False

            if new:
                # saving session data
                for idx, session in enumerate(sessions):
                    logger.info(
                        os.path.join(
                            project_path,
                            "results",
                            session,
                            "",
                            model_name,
                            seg + "-" + str(n_clusters),
                            "",
                        )
                    )
                    if not os.path.exists(
                        os.path.join(
                            project_path,
                            "results",
                            session,
                            model_name,
                            seg + "-" + str(n_clusters),
                            "",
                        )
                    ):
                        try:
                            os.mkdir(
                                os.path.join(
                                    project_path,
                                    "results",
                                    session,
                                    "",
                                    model_name,
                                    seg + "-" + str(n_clusters),
                                    "",
                                )
                            )
                        except OSError as error:
                            logger.error(error)

                    save_data = os.path.join(
                        str(project_path),
                        "results",
                        session,
                        model_name,
                        seg + "-" + str(n_clusters),
                        "",
                    )
                    np.save(
                        os.path.join(
                            save_data,
                            str(n_clusters) + "_" + seg + "_label_" + session,
                        ),
                        labels[idx],
                    )
                    if seg == "kmeans":
                        np.save(
                            os.path.join(save_data, "cluster_center_" + session),
                            cluster_center[idx],
                        )
                    np.save(
                        os.path.join(save_data, "latent_vector_" + session),
                        latent_vectors[idx],
                    )
                    np.save(
                        os.path.join(save_data, "motif_usage_" + session),
                        motif_usages[idx],
                    )

                logger.info(
                    "You succesfully extracted motifs with VAME! From here, you can proceed running vame.motif_videos()"
                )
                # "to get an idea of the behavior captured by VAME. This will leave you with short snippets of certain movements."
                # "To get the full picture of the spatiotemporal dynamic we recommend applying our community approach afterwards.")

    except Exception as e:
        logger.exception(f"An error occurred during pose segmentation: {e}")
    finally:
        logger_config.remove_file_handler()



import sys

sys.dont_write_bytecode = True

from vame.analysis.pose_segmentation import segment_session
from vame.analysis.videowriter import motif_videos, community_videos
from vame.analysis.community_analysis import community
from vame.analysis.generative_functions import generative_model
from vame.analysis.gif_creator import gif



from pathlib import Path
import numpy as np
from scipy.stats import iqr

from vame.logging.logger import VameLogger
from vame.io.load_poses import read_pose_estimation_file


logger_config = VameLogger(__name__)
logger = logger_config.logger


def lowconf_cleaning(
    config: dict,
    read_from_variable: str = "position_processed",
    save_to_variable: str = "position_processed",
) -> None:
    """
    Clean the low confidence data points from the dataset. Processes position data by:
     - setting low-confidence points to NaN
     - interpolating NaN points

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    read_from_variable : str, optional
        Variable to read from the dataset.
    save_to_variable : str, optional
        Variable to save the cleaned data to.

    Returns
    -------
    None
    """
    project_path = config["project_path"]
    sessions = config["session_names"]
    pose_confidence = config["pose_confidence"]
    logger.info(f"Cleaning low confidence data points. Confidence threshold: {pose_confidence}")

    for i, session in enumerate(sessions):
        logger.info(f"Session: {session}")
        # Read session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)

        position = ds[read_from_variable].values
        cleaned_position = np.empty_like(position)
        confidence = ds["confidence"].values

        perc_interp_points = np.zeros((position.shape[1], position.shape[2], position.shape[3]))
        for individual in range(position.shape[1]):
            for keypoint in range(position.shape[2]):
                conf_series = confidence[:, individual, keypoint]
                for space in range(position.shape[3]):
                    # Set low-confidence positions to NaN
                    nan_mask = conf_series < pose_confidence
                    series = np.copy(position[:, individual, keypoint, space])
                    series[nan_mask] = np.nan

                    # Update nan_mask because the series might come with NaN values previously
                    nan_mask = np.isnan(series)

                    perc_interp_points[individual, keypoint, space] = 100 * np.sum(nan_mask) / len(nan_mask)

                    # Interpolate NaN values
                    if not nan_mask.all():
                        series[nan_mask] = np.interp(
                            np.flatnonzero(nan_mask),
                            np.flatnonzero(~nan_mask),
                            series[~nan_mask],
                        )

                    # Update the position array
                    cleaned_position[:, individual, keypoint, space] = series

        # Update the dataset with the cleaned position values
        ds[save_to_variable] = (ds[read_from_variable].dims, cleaned_position)
        ds.attrs.update({"processed_confidence": True})

        ds["percentage_low_confidence"] = (["individual", "keypoint", "space"], perc_interp_points)

        # Save the cleaned dataset to file
        cleaned_file_path = Path(project_path) / "data" / "processed" / f"{session}_processed.nc"
        ds.to_netcdf(
            path=cleaned_file_path,
            engine="scipy",
        )


def outlier_cleaning(
    config: dict,
    read_from_variable: str = "position_processed",
    save_to_variable: str = "position_processed",
) -> None:
    """
    Clean the outliers from the dataset. Processes position data by:
     - setting outlier points to NaN
     - interpolating NaN points

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    read_from_variable : str, optional
        Variable to read from the dataset.
    save_to_variable : str, optional
        Variable to save the cleaned data to.

    Returns
    -------
    None
    """
    logger.info("Cleaning outliers with Z-score transformation and IQR cutoff.")
    project_path = config["project_path"]
    sessions = config["session_names"]

    for i, session in enumerate(sessions):
        logger.info(f"Session: {session}")
        # Read raw session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)

        position = np.copy(ds[read_from_variable].values)
        cleaned_position = np.copy(position)

        perc_interp_points = np.zeros((position.shape[1], position.shape[2], position.shape[3]))

        for individual in range(position.shape[1]):
            for keypoint in range(position.shape[2]):
                for space in range(position.shape[3]):
                    series = np.copy(position[:, individual, keypoint, space])

                    # Check if all values are zero, then skip
                    if np.all(series == 0):
                        continue

                    # Calculate Z-score
                    z_series = (series - np.nanmean(series)) / np.nanstd(series)

                    # Set outlier positions to NaN, based on IQR cutoff
                    if config["robust"]:
                        iqr_factor = config["iqr_factor"]
                        iqr_val = iqr(z_series)
                        outlier_mask = np.abs(z_series) > iqr_factor * iqr_val
                        z_series[outlier_mask] = np.nan
                        perc_interp_points[individual, keypoint, space] = (
                            100 * np.sum(outlier_mask) / len(outlier_mask)
                        )

                        # Interpolate NaN values
                        if not outlier_mask.all():
                            z_series[outlier_mask] = np.interp(
                                np.flatnonzero(outlier_mask),
                                np.flatnonzero(~outlier_mask),
                                z_series[~outlier_mask],
                            )

                        # Redo the z-score to remove the bias of the now-removed outliers
                        z_series = (z_series - np.nanmean(z_series)) / np.nanstd(z_series)

                    # Update the processed position array
                    cleaned_position[:, individual, keypoint, space] = z_series

        # Update the dataset with the cleaned position values
        ds[save_to_variable] = (ds[read_from_variable].dims, cleaned_position)
        ds.attrs.update({"processed_outliers": True})

        ds["percentage_iqr_outliers"] = (["individual", "keypoint", "space"], perc_interp_points)

        # Save the cleaned dataset to file
        cleaned_file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        ds.to_netcdf(
            path=cleaned_file_path,
            engine="scipy",
        )



from typing import List, Optional, Literal
from pathlib import Path
import matplotlib.pyplot as plt
import xarray as xr

import vame
from vame.util.auxiliary import read_states
from vame.io.load_poses import load_vame_dataset
from vame.visualization.motif import visualize_motif_tree
from vame.visualization.umap import visualize_umap
from vame.visualization.preprocessing import (
    visualize_preprocessing_scatter,
    visualize_preprocessing_timeseries,
)
from vame.visualization.model import plot_loss
from vame.logging.logger import VameLogger


logger_config = VameLogger(__name__)
logger = logger_config.logger


class VAMEPipeline:
    """VAME pipeline class."""

    def __init__(
        self,
        project_name: str,
        videos: List[str],
        poses_estimations: List[str],
        source_software: Literal["DeepLabCut", "SLEAP", "LightningPose"],
        working_directory: str = ".",
        video_type: str = ".mp4",
        fps: int | None = None,
        copy_videos: bool = False,
        paths_to_pose_nwb_series_data: Optional[str] = None,
        config_kwargs: Optional[dict] = None,
    ) -> None:
        """
        Initializes the VAME pipeline.

        Parameters
        ----------
        project_name : str
            Project name.
        videos : List[str]
            List of video files.
        poses_estimations : List[str]
            List of pose estimation files.
        source_software : Literal["DeepLabCut", "SLEAP", "LightningPose"]
            Source software used for pose estimation.
        working_directory : str, optional
            Working directory, by default ".".
        video_type : str, optional
            Video file type, by default ".mp4".
        fps : int, optional
            Sampling rate of the videos. If not passed, it will be estimated from the video file. By default None.
        copy_videos : bool, optional
            Copy videos, by default False.
        paths_to_pose_nwb_series_data : Optional[str], optional
            Path to pose NWB series data, by default None.
        config_kwargs : Optional[dict], optional
            Additional configuration keyword arguments, by default None.

        Returns
        -------
        None
        """
        self.config_path, self.config = vame.init_new_project(
            project_name=project_name,
            poses_estimations=poses_estimations,
            source_software=source_software,
            working_directory=working_directory,
            videos=videos,
            video_type=video_type,
            fps=fps,
            copy_videos=copy_videos,
            paths_to_pose_nwb_series_data=paths_to_pose_nwb_series_data,
            config_kwargs=config_kwargs,
        )

    def get_states(self, summary: bool = True) -> dict:
        """
        Returns the pipeline states.

        Returns
        -------
        dict
            Pipeline states.
        """
        states = read_states(self.config)
        if summary and states:
            logger.info("Pipeline states:")
            for key, value in states.items():
                logger.info(f"{key}: {value.get('execution_state', 'Not executed')}")
        return states

    def get_sessions(self) -> List[str]:
        """
        Returns a list of session names.

        Returns
        -------
        List[str]
            Session names.
        """
        return self.config["session_names"]

    def get_raw_datasets(self) -> xr.Dataset:
        """
        Returns a xarray dataset which combines all the raw data from the project.

        Returns
        -------
        dss : xarray.Dataset
            Combined raw dataset.
        """
        sessions = self.get_sessions()
        datasets = list()
        attributes = list()
        for session in sessions:
            ds_path = Path(self.config["project_path"]) / "data" / "raw" / f"{session}.nc"
            ds = load_vame_dataset(ds_path=ds_path)
            ds = ds.expand_dims({"session": [session]})
            datasets.append(ds)
            attributes.append(ds.attrs)
        dss = xr.concat(datasets, dim="session")
        dss_attrs = {}
        for d in attributes:
            for key, value in d.items():
                dss_attrs.setdefault(key, []).append(value)
        for key, values in dss_attrs.items():
            unique_values = unique_in_order(values)  # Maintain order of unique values
            dss_attrs[key] = unique_values[0] if len(unique_values) == 1 else unique_values
        for key, value in dss_attrs.items():
            dss.attrs[key] = value
        return dss

    def preprocessing(
        self,
        centered_reference_keypoint: str = "snout",
        orientation_reference_keypoint: str = "tailbase",
    ) -> None:
        """
        Preprocesses the data.

        Parameters
        ----------
        centered_reference_keypoint : str, optional
            Key point to center the data, by default "snout".
        orientation_reference_keypoint : str, optional
            Key point to orient the data, by default "tailbase".

        Returns
        -------
        None
        """
        self.centered_reference_keypoint = centered_reference_keypoint
        self.orientation_reference_keypoint = orientation_reference_keypoint
        vame.preprocessing(
            config=self.config,
            centered_reference_keypoint=centered_reference_keypoint,
            orientation_reference_keypoint=orientation_reference_keypoint,
        )

    def create_training_set(self) -> None:
        """
        Creates the training set.

        Returns
        -------
        None
        """
        vame.create_trainset(config=self.config)

    def train_model(self) -> None:
        """
        Trains the model.

        Returns
        -------
        None
        """
        vame.train_model(config=self.config)

    def evaluate_model(self) -> None:
        """
        Evaluates the model.

        Returns
        -------
        None
        """
        vame.evaluate_model(config=self.config)

    def run_segmentation(self) -> None:
        """
        Runs the pose estimation segmentation into motifs.

        Returns
        -------
        None
        """
        vame.segment_session(config=self.config)

    def run_community_clustering(self) -> None:
        """
        Runs the community clustering.

        Returns
        -------
        None
        """
        vame.community(
            config=self.config,
            segmentation_algorithm="hmm",
            cohort=True,
            cut_tree=2,
        )

    def generate_motif_videos(
        self,
        video_type: str = ".mp4",
        segmentation_algorithm: Literal["hmm", "kmeans"] = "hmm",
    ) -> None:
        """
        Generates motif videos.

        Parameters
        ----------
        video_type : str, optional
            Video type, by default ".mp4".
        segmentation_algorithm : Literal["hmm", "kmeans"], optional
            Segmentation algorithm, by default "hmm".

        Returns
        -------
        None
        """
        vame.motif_videos(
            config=self.config,
            video_type=video_type,
            segmentation_algorithm=segmentation_algorithm,
        )

    def generate_community_videos(
        self,
        video_type: str = ".mp4",
        segmentation_algorithm: Literal["hmm", "kmeans"] = "hmm",
    ) -> None:
        """
        Generates community videos.

        Parameters
        ----------
        video_type : str, optional
            Video type, by default ".mp4".
        segmentation_algorithm : Literal["hmm", "kmeans"], optional
            Segmentation algorithm, by default "hmm".

        Returns
        -------
        None
        """
        vame.community_videos(
            config=self.config,
            video_type=video_type,
            segmentation_algorithm=segmentation_algorithm,
        )

    def generate_videos(
        self,
        video_type: str = ".mp4",
        segmentation_algorithm: Literal["hmm", "kmeans"] = "hmm",
    ) -> None:
        """
        Generates motif and community videos.

        Parameters
        ----------
        video_type : str, optional
            Video type, by default ".mp4".
        segmentation_algorithm : Literal["hmm", "kmeans"], optional
            Segmentation algorithm, by default "hmm".

        Returns
        -------
        None
        """
        self.generate_motif_videos(
            video_type=video_type,
            segmentation_algorithm=segmentation_algorithm,
        )
        self.generate_community_videos(
            video_type=video_type,
            segmentation_algorithm=segmentation_algorithm,
        )

    def visualize_preprocessing(
        self,
        scatter: bool = True,
        timeseries: bool = True,
        show_figure: bool = False,
        save_to_file: bool = True,
    ) -> None:
        """
        Visualizes the preprocessing results.

        Parameters
        ----------
        scatter : bool, optional
            Visualize scatter plot, by default True.
        timeseries : bool, optional
            Visualize timeseries plot, by default True.
        show_figure : bool, optional
            Show the figure, by default False.
        save_to_file : bool, optional
            Save the figure to file, by default True.

        Returns
        -------
        None
        """
        if scatter:
            visualize_preprocessing_scatter(
                config=self.config,
                show_figure=show_figure,
                save_to_file=save_to_file,
            )
        if timeseries:
            visualize_preprocessing_timeseries(
                config=self.config,
                show_figure=show_figure,
                save_to_file=save_to_file,
            )

    def visualize_model_losses(
        self,
        save_to_file: bool = True,
        show_figure: bool = True,
    ) -> None:
        """
        Visualizes the model losses.

        Parameters
        ----------
        save_to_file : bool, optional
            Save the figure to file, by default False.
        show_figure : bool, optional
            Show the figure, by default True.

        Returns
        -------
        None
        """
        plot_loss(
            cfg=self.config,
            model_name="VAME",
            save_to_file=save_to_file,
            show_figure=show_figure,
        )

    def visualize_motif_tree(
        self,
        segmentation_algorithm: Literal["hmm", "kmeans"],
    ) -> None:
        """
        Visualizes the motif tree.

        Parameters
        ----------
        segmentation_algorithm : Literal["hmm", "kmeans"]
            Segmentation algorithm.

        Returns
        -------
        None
        """
        visualize_motif_tree(
            config=self.config,
            segmentation_algorithm=segmentation_algorithm,
        )

    def visualize_umap(
        self,
        label: Literal["community", "motif"] = "community",
        segmentation_algorithm: Literal["hmm", "kmeans"] = "hmm",
        show_figure: bool = False,
        save_to_file: bool = True,
    ) -> None:
        """
        Visualizes the UMAP plot.

        Parameters
        ----------
        label : Literal["community", "motif"], optional
            Label to visualize, by default "community".
        segmentation_algorithm : Literal["hmm", "kmeans"], optional
            Segmentation algorithm, by default "hmm".

        Returns
        -------
        None
        """
        visualize_umap(
            config=self.config,
            label=label,
            segmentation_algorithm=segmentation_algorithm,
        )

    def report(
        self,
        segmentation_algorithm: Literal["hmm", "kmeans"] = "hmm",
    ) -> None:
        """
        Generates the project report.

        Parameters
        ----------
        segmentation_algorithm : Literal["hmm", "kmeans"], optional
            Segmentation algorithm, by default "hmm".

        Returns
        -------
        None
        """
        vame.report(
            config=self.config,
            segmentation_algorithm=segmentation_algorithm,
        )

    def run_pipeline(
        self,
        from_step: int = 0,
        preprocessing_kwargs: dict = {},
    ) -> None:
        """
        Runs the pipeline.

        Parameters
        ----------
        from_step : int, optional
            Start from step, by default 0.
        preprocessing_kwargs : dict, optional
            Preprocessing keyword arguments, by default {}.

        Returns
        -------
        None
        """
        if from_step == 0:
            self.preprocessing(**preprocessing_kwargs)
        if from_step <= 1:
            self.create_training_set()
        if from_step <= 2:
            self.train_model()
        if from_step <= 3:
            self.evaluate_model()
        if from_step <= 4:
            self.run_segmentation()
        if from_step <= 5:
            self.run_community_clustering()


def unique_in_order(sequence):
    seen = set()
    result = []
    for item in sequence:
        if item not in seen:
            result.append(item)
            seen.add(item)
    return result



import os
from typing import Optional
import numpy as np
from matplotlib import pyplot as plt
import torch
import torch.utils.data as Data

from vame.model.rnn_vae import RNN_VAE


use_gpu = torch.cuda.is_available()
if use_gpu:
    pass
else:
    torch.device("cpu")


def plot_reconstruction(
    filepath: str,
    test_loader: Data.DataLoader,
    seq_len_half: int,
    model: RNN_VAE,
    model_name: str,
    FUTURE_DECODER: bool,
    FUTURE_STEPS: int,
    suffix: Optional[str] = None,
    show_figure: bool = False,
) -> None:
    """
    Plot the reconstruction and future prediction of the input sequence.
    Saves the plot to:
    - project_name/
        - model/
            - evaluate/
                - Reconstruction_model_name.png

    Parameters
    ----------
    filepath : str
        Path to save the plot.
    test_loader : Data.DataLoader
        DataLoader for the test dataset.
    seq_len_half : int
        Half of the temporal window size.
    model : RNN_VAE
        Trained VAE model.
    model_name : str
        Name of the model.
    FUTURE_DECODER : bool
        Flag indicating whether the model has a future prediction decoder.
    FUTURE_STEPS : int
        Number of future steps to predict.
    suffix : str, optional
        Suffix for the saved plot filename. Defaults to None.
    show_figure : bool, optional
        Flag indicating whether to show the plot. Defaults to False.

    Returns
    -------
    None
    """
    # x = test_loader.__iter__().next()
    dataiter = iter(test_loader)
    x = next(dataiter)
    x = x.permute(0, 2, 1)
    if use_gpu:
        data = x[:, :seq_len_half, :].type("torch.FloatTensor").cuda()
        data_fut = x[:, seq_len_half : seq_len_half + FUTURE_STEPS, :].type("torch.FloatTensor").cuda()
    else:
        data = x[:, :seq_len_half, :].type("torch.FloatTensor").to()
        data_fut = x[:, seq_len_half : seq_len_half + FUTURE_STEPS, :].type("torch.FloatTensor").to()
    if FUTURE_DECODER:
        x_tilde, future, latent, mu, logvar = model(data)

        fut_orig = data_fut.cpu()
        fut_orig = fut_orig.data.numpy()
        fut = future.cpu()
        fut = fut.detach().numpy()

    else:
        x_tilde, latent, mu, logvar = model(data)

    data_orig = data.cpu()
    data_orig = data_orig.data.numpy()
    data_tilde = x_tilde.cpu()
    data_tilde = data_tilde.detach().numpy()

    if FUTURE_DECODER:
        fig, axs = plt.subplots(2, 5)
        fig.suptitle("Reconstruction [top] and future prediction [bottom] of input sequence")
        for i in range(5):
            axs[0, i].plot(data_orig[i, ...], color="k", label="Sequence Data")
            axs[0, i].plot(
                data_tilde[i, ...],
                color="r",
                linestyle="dashed",
                label="Sequence Reconstruction",
            )
            axs[1, i].plot(fut_orig[i, ...], color="k")
            axs[1, i].plot(fut[i, ...], color="r", linestyle="dashed")
        axs[0, 0].set(xlabel="time steps", ylabel="reconstruction")
        axs[1, 0].set(xlabel="time steps", ylabel="predction")
        fig.savefig(os.path.join(filepath, "evaluate", "Future_Reconstruction.png"))
    else:
        fig, ax1 = plt.subplots(1, 5)
        for i in range(5):
            fig.suptitle("Reconstruction of input sequence")
            ax1[i].plot(data_orig[i, ...], color="k", label="Sequence Data")
            ax1[i].plot(
                data_tilde[i, ...],
                color="r",
                linestyle="dashed",
                label="Sequence Reconstruction",
            )
        fig.tight_layout()
        if not suffix:
            fig.savefig(
                os.path.join(filepath, "evaluate", "Reconstruction_" + model_name + ".png"),
                bbox_inches="tight",
            )
        elif suffix:
            fig.savefig(
                os.path.join(
                    filepath,
                    "evaluate",
                    "Reconstruction_" + model_name + "_" + suffix + ".png",
                ),
                bbox_inches="tight",
            )

    if show_figure:
        plt.show()
    else:
        plt.close(fig)


def plot_loss(
    cfg: dict,
    model_name: str,
    save_to_file: bool = False,
    show_figure: bool = True,
) -> None:
    """
    Plot the losses of the trained model.
    Saves the plot to:
    - project_name/
        - model/
            - evaluate/
                - MSE-and-KL-Loss_model_name.png

    Parameters
    ----------
    cfg : dict
        Configuration dictionary.
    model_name : str
        Name of the model.
    save_to_file : bool, optional
        Flag indicating whether to save the plot. Defaults to False.
    show_figure : bool, optional
        Flag indicating whether to show the plot. Defaults to True.

    Returns
    -------
    None
    """
    basepath = os.path.join(cfg["project_path"], "model", "model_losses")
    train_loss = np.load(os.path.join(basepath, "train_losses_" + model_name + ".npy"))
    test_loss = np.load(os.path.join(basepath, "test_losses_" + model_name + ".npy"))
    mse_loss_train = np.load(os.path.join(basepath, "mse_train_losses_" + model_name + ".npy"))
    mse_loss_test = np.load(os.path.join(basepath, "mse_test_losses_" + model_name + ".npy"))
    km_losses = np.load(os.path.join(basepath, "kmeans_losses_" + model_name + ".npy"))
    kl_loss = np.load(os.path.join(basepath, "kl_losses_" + model_name + ".npy"))
    fut_loss = np.load(os.path.join(basepath, "fut_losses_" + model_name + ".npy"))

    fig, (ax1) = plt.subplots(1, 1)
    fig.suptitle(f"Losses of model: {model_name}")
    ax1.set(xlabel="Epochs", ylabel="loss [log-scale]")
    ax1.set_yscale("log")
    ax1.plot(train_loss, label="Train-Loss")
    ax1.plot(test_loss, label="Test-Loss")
    ax1.plot(mse_loss_train, label="MSE-Train-Loss")
    ax1.plot(mse_loss_test, label="MSE-Test-Loss")
    ax1.plot(km_losses, label="KMeans-Loss")
    ax1.plot(kl_loss, label="KL-Loss")
    ax1.plot(fut_loss, label="Prediction-Loss")
    ax1.legend()

    if save_to_file:
        evaluate_path = os.path.join(cfg["project_path"], "model", "evaluate")
        fig.savefig(os.path.join(evaluate_path, "MSE-and-KL-Loss" + model_name + ".png"))

    if show_figure:
        plt.show()
    else:
        plt.close(fig)



from pydantic import BaseModel, Field, ConfigDict
from typing import List, Optional, Literal
from datetime import datetime, timezone
from enum import Enum


class SegmentationAlgorithms(str, Enum):
    hmm = "hmm"
    kmeans = "kmeans"

    class Config:
        use_enum_values = True


class ProjectSchema(BaseModel):
    # Project parameters
    project_name: str = Field(
        ...,
        title="Project name",
    )
    creation_datetime: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat(timespec="seconds"),
        title="Creation datetime",
    )
    model_name: str = Field(
        default="VAME",
        title="Model name",
    )
    n_clusters: int = Field(
        default=15,
        title="Number of clusters",
    )
    pose_confidence: float = Field(
        default=0.99,
        title="Pose confidence",
    )
    project_path: str = Field(
        ...,
        title="Project path",
    )
    session_names: List[str] = Field(
        ...,
        title="Session names",
    )
    pose_estimation_filetype: Literal["csv", "nwb", "slp", "h5"] = Field(
        title="Pose estimation filetype",
    )
    paths_to_pose_nwb_series_data: Optional[List[str]] = Field(
        title="Paths to pose series data in nwb files",
        default=None,
    )

    # Data
    all_data: str = Field(
        default="yes",
        title="All data",
    )
    egocentric_data: bool = Field(
        default=False,
        title="Egocentric data",
    )
    robust: bool = Field(
        default=True,
        title="Robust data",
    )
    iqr_factor: int = Field(
        default=4,
        title="IQR factor",
    )
    axis: str = Field(
        default="None",
        title="Axis",
    )
    savgol_filter: bool = Field(
        default=True,
        title="Savgol filter",
    )
    savgol_length: int = Field(
        default=5,
        title="Savgol length",
    )
    savgol_order: int = Field(
        default=2,
        title="Savgol order",
    )
    test_fraction: float = Field(
        default=0.1,
        title="Test fraction",
    )

    # RNN model general hyperparameters
    pretrained_model: str = Field(
        default="None",
        title="Pretrained model",
    )
    pretrained_weights: bool = Field(
        default=False,
        title="Pretrained weights",
    )
    num_features: int = Field(
        default=12,
        title="Number of features",
    )
    batch_size: int = Field(
        default=256,
        title="Batch size",
    )
    max_epochs: int = Field(
        default=500,
        title="Max epochs",
    )
    model_snapshot: int = Field(
        default=50,
        title="Model snapshot",
    )
    model_convergence: int = Field(
        default=50,
        title="Model convergence",
    )
    transition_function: str = Field(
        default="GRU",
        title="Transition function",
    )
    beta: float = Field(
        default=1,
        title="Beta",
    )
    beta_norm: bool = Field(
        default=False,
        title="Beta normalization",
    )
    zdims: int = Field(
        default=30,
        title="Zdims",
    )
    learning_rate: float = Field(
        default=5e-4,
        title="Learning rate",
    )
    time_window: int = Field(
        default=30,
        title="Time window",
    )
    prediction_decoder: int = Field(
        default=1,
        title="Prediction decoder",
    )
    prediction_steps: int = Field(
        default=15,
        title="Prediction steps",
    )
    noise: bool = Field(
        default=False,
        title="Noise",
    )
    scheduler: int = Field(
        default=1,
        title="Scheduler",
    )
    scheduler_step_size: int = Field(
        default=100,
        title="Scheduler step size",
    )
    scheduler_gamma: float = Field(
        default=0.2,
        title="Scheduler gamma",
    )
    scheduler_threshold: Optional[float] = Field(
        default=None,
        title="Scheduler threshold",
    )
    softplus: bool = Field(
        default=False,
        title="Softplus",
    )

    # Segmentation
    segmentation_algorithms: List[SegmentationAlgorithms] = Field(
        title="Segmentation algorithms",
        default_factory=lambda: [
            SegmentationAlgorithms.hmm,
            SegmentationAlgorithms.kmeans,
        ],
    )
    hmm_trained: bool = Field(
        default=False,
        title="HMM trained",
    )
    load_data: str = Field(
        default="-PE-seq-clean",
        title="Load data",
    )
    individual_segmentation: bool = Field(
        default=False,
        title="Individual segmentation",
    )
    random_state_kmeans: int = Field(
        default=42,
        title="Random state kmeans",
    )
    n_init_kmeans: int = Field(
        default=15,
        title="N init kmeans",
    )

    # Video writer:
    length_of_motif_video: int = Field(
        default=1000,
        title="Length of motif video",
    )

    # UMAP parameter:
    min_dist: float = Field(
        default=0.1,
        title="Min dist",
    )
    n_neighbors: int = Field(
        default=200,
        title="N neighbors",
    )
    random_state: int = Field(
        default=42,
        title="Random state",
    )
    num_points: int = Field(
        default=30000,
        title="Num points",
    )

    # RNN encoder hyperparameter:
    hidden_size_layer_1: int = Field(
        default=256,
        title="Hidden size layer 1",
    )
    hidden_size_layer_2: int = Field(
        default=256,
        title="Hidden size layer 2",
    )
    dropout_encoder: float = Field(
        default=0,
        title="Dropout encoder",
    )

    # RNN reconstruction hyperparameter:
    hidden_size_rec: int = Field(
        default=256,
        title="Hidden size rec",
    )
    dropout_rec: float = Field(
        default=0,
        title="Dropout rec",
    )
    n_layers: int = Field(
        default=1,
        title="N layers",
    )

    # RNN prediction hyperparameter:
    hidden_size_pred: int = Field(
        default=256,
        title="Hidden size pred",
    )
    dropout_pred: float = Field(
        default=0,
        title="Dropout pred",
    )

    # RNN loss hyperparameter:
    mse_reconstruction_reduction: str = Field(
        default="sum",
        title="MSE reconstruction reduction",
    )
    mse_prediction_reduction: str = Field(
        default="sum",
        title="MSE prediction reduction",
    )
    kmeans_loss: int = Field(
        default=30,
        title="Kmeans loss",
    )
    kmeans_lambda: float = Field(
        default=0.1,
        title="Kmeans lambda",
    )
    anneal_function: str = Field(
        default="linear",
        title="Anneal function",
    )
    kl_start: int = Field(
        default=2,
        title="KL start",
    )
    annealtime: int = Field(
        default=4,
        title="Annealtime",
    )

    model_config: ConfigDict = ConfigDict(
        protected_namespaces=(),
        use_enum_values=True,
    )



# import os
# import cv2 as cv
# import numpy as np
# import pandas as pd
# import tqdm
# from typing import Tuple, List, Union
# from pathlib import Path

# from vame.logging.logger import VameLogger, TqdmToLogger
# from vame.util.auxiliary import read_config
# from vame.schemas.states import EgocentricAlignmentFunctionSchema, save_state
# from vame.schemas.project import PoseEstimationFiletype
# from vame.io.load_poses import read_pose_estimation_file
# from vame.util.data_manipulation import (
#     interpol_first_rows_nans,
#     crop_and_flip_legacy,
#     background,
# )
# from vame.video import get_video_frame_rate


# logger_config = VameLogger(__name__)
# logger = logger_config.logger


# def align_mouse_legacy(
#     project_path: str,
#     session: str,
#     video_format: str,
#     crop_size: Tuple[int, int],
#     pose_list: List[np.ndarray],
#     pose_ref_index: Tuple[int, int],
#     confidence: float,
#     pose_flip_ref: Tuple[int, int],
#     bg: np.ndarray,
#     frame_count: int,
#     use_video: bool = True,
#     tqdm_stream: Union[TqdmToLogger, None] = None,
# ) -> Tuple[List[np.ndarray], List[List[np.ndarray]], np.ndarray]:
#     """
#     Align the mouse in the video frames.

#     Parameters
#     ----------
#     project_path : str
#         Path to the project directory.
#     session : str
#         Name of the session.
#     video_format : str
#         Format of the video file.
#     crop_size : Tuple[int, int]
#         Size to crop the video frames.
#     pose_list : List[np.ndarray]
#         List of pose coordinates.
#     pose_ref_index : Tuple[int, int]
#         Pose reference indices.
#     confidence : float
#         Pose confidence threshold.
#     pose_flip_ref : Tuple[int, int]
#         Reference indices for flipping.
#     bg : np.ndarray
#         Background image.
#     frame_count : int
#         Number of frames to align.
#     use_video : bool, optional
#         bool if video should be cropped or DLC points only. Defaults to True.
#     tqdm_stream : Union[TqdmToLogger, None], optional
#         Tqdm stream to log the progress. Defaults to None.

#     Returns
#     -------
#     Tuple[List[np.ndarray], List[List[np.ndarray]], np.ndarray]
#         List of aligned images, list of aligned DLC points, and aligned time series data.
#     """
#     images = []
#     points = []
#     for i in pose_list:
#         for j in i:
#             if j[2] <= confidence:
#                 j[0], j[1] = np.nan, np.nan

#     for i in pose_list:
#         i = interpol_first_rows_nans(i)

#     if use_video:
#         video_path = str(
#             os.path.join(
#                 project_path,
#                 "data",
#                 "raw",
#                 session + video_format,
#             )
#         )
#         capture = cv.VideoCapture(video_path)
#         if not capture.isOpened():
#             raise Exception(f"Unable to open video file: {video_path}")

#     for idx in tqdm.tqdm(
#         range(frame_count),
#         disable=not True,
#         file=tqdm_stream,
#         desc="Align frames",
#     ):
#         if use_video:
#             # Read frame
#             try:
#                 ret, frame = capture.read()
#                 frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
#                 frame = frame - bg
#                 frame[frame <= 0] = 0
#             except Exception:
#                 logger.info("Couldn't find a frame in capture.read(). #Frame: %d" % idx)
#                 continue
#         else:
#             frame = np.zeros((1, 1))

#         # Read coordinates and add border
#         pose_list_bordered = []

#         for i in pose_list:
#             pose_list_bordered.append((int(i[idx][0] + crop_size[0]), int(i[idx][1] + crop_size[1])))

#         img = cv.copyMakeBorder(
#             frame,
#             crop_size[1],
#             crop_size[1],
#             crop_size[0],
#             crop_size[0],
#             cv.BORDER_CONSTANT,
#             0,
#         )

#         coord_center = []
#         punkte = []

#         for i in pose_ref_index:
#             coord = []
#             # changed from pose_list_bordered[i][0] 2/28/2024 PN
#             coord.append(pose_list_bordered[i][0])
#             # changed from pose_list_bordered[i][1] 2/28/2024 PN
#             coord.append(pose_list_bordered[i][1])
#             punkte.append(coord)

#         # coord_center.append(pose_list_bordered[5][0]-5)
#         # coord_center.append(pose_list_bordered[5][0]+5)
#         # coord_center = [coord_center]
#         punkte = [punkte]

#         # coord_center = np.asarray(coord_center)
#         punkte = np.asarray(punkte)

#         # calculate minimal rectangle around snout and tail
#         rect = cv.minAreaRect(punkte)
#         # rect_belly = cv.minAreaRect(coord_center)
#         # center_belly, size_belly, theta_belly = rect_belly
#         # change size in rect tuple structure to be equal to crop_size
#         lst = list(rect)
#         lst[1] = crop_size
#         # lst[0] = center_belly
#         rect = tuple(lst)
#         center, size, theta = rect

#         # crop image
#         out, shifted_points = crop_and_flip_legacy(
#             rect,
#             img,
#             pose_list_bordered,
#             pose_flip_ref,
#         )

#         if use_video:  # for memory optimization, just save images when video is used.
#             images.append(out)
#         points.append(shifted_points)

#     if use_video:
#         capture.release()

#     time_series = np.zeros((len(pose_list) * 2, frame_count))
#     for i in range(frame_count):
#         idx = 0
#         for j in range(len(pose_list)):
#             time_series[idx : idx + 2, i] = points[i][j]
#             idx += 2

#     return images, points, time_series


# def alignment_legacy(
#     project_path: str,
#     session: str,
#     pose_ref_index: Tuple[int, int],
#     video_format: str,
#     crop_size: Tuple[int, int],
#     confidence: float,
#     pose_estimation_filetype: PoseEstimationFiletype,
#     path_to_pose_nwb_series_data: Union[str, None] = None,
#     use_video: bool = False,
#     tqdm_stream: Union[TqdmToLogger, None] = None,
# ) -> Tuple[np.ndarray, List[np.ndarray]]:
#     """
#     Perform alignment of egocentric data.

#     Parameters
#     ----------
#     project_path : str
#         Path to the project directory.
#     session : str
#         Name of the session.
#     pose_ref_index : List[int]
#         Pose reference indices.
#     video_format : str
#         Format of the video file.
#     crop_size : Tuple[int, int]
#         Size to crop the video frames.
#     confidence : float
#         Pose confidence threshold.
#     pose_estimation_filetype : PoseEstimationFiletype
#         Pose estimation file type. Can be .csv or .nwb.
#     path_to_pose_nwb_series_data : Union[str, None], optional
#         Path to the pose series data in nwb files. Defaults to None.
#     use_video : bool, optional
#         Whether to use video for alignment. Defaults to False.
#     tqdm_stream : Union[TqdmToLogger, None], optional
#         Tqdm stream to log the progress. Defaults to None.

#     Returns
#     -------
#     Tuple[np.ndarray, List[np.ndarray]]
#         Aligned time series data and list of aligned frames.
#     """
#     # read out data
#     file_path = str(Path(project_path) / "data" / "raw" / f"{session}.nc")
#     data, data_mat, _ = read_pose_estimation_file(
#         file_path=file_path,
#         file_type=pose_estimation_filetype,
#         path_to_pose_nwb_series_data=path_to_pose_nwb_series_data,
#     )

#     # get the coordinates for alignment from data table
#     # pose_list dimensions: (num_body_parts, num_frames, 3)
#     pose_list = []
#     for i in range(int(data_mat.shape[1] / 3)):
#         pose_list.append(data_mat[:, i * 3 : (i + 1) * 3])

#     # list of reference coordinate indices for alignment
#     # 0: snout, 1: forehand_left, 2: forehand_right,
#     # 3: hindleft, 4: hindright, 5: tail
#     # list of 2 reference coordinate indices for avoiding flipping
#     pose_flip_ref = pose_ref_index

#     if use_video:
#         # compute background
#         video_path = str(
#             os.path.join(
#                 project_path,
#                 "data",
#                 "raw",
#                 session + video_format,
#             )
#         )
#         bg = background(
#             project_path=project_path,
#             session=session,
#             video_path=video_path,
#             save_background=False,
#         )
#         frame_count = get_video_frame_rate(video_path)
#     else:
#         bg = 0
#         # Change this to an abitrary number if you first want to test the code
#         frame_count = len(data)

#     frames, n, time_series = align_mouse_legacy(
#         project_path=project_path,
#         session=session,
#         video_format=video_format,
#         crop_size=crop_size,
#         pose_list=pose_list,
#         pose_ref_index=pose_ref_index,
#         confidence=confidence,
#         pose_flip_ref=pose_flip_ref,
#         bg=bg,
#         frame_count=frame_count,
#         use_video=use_video,
#         tqdm_stream=tqdm_stream,
#     )

#     return time_series, frames


# # @save_state(model=EgocentricAlignmentFunctionSchema)
# def egocentric_alignment_legacy(
#     config: str,
#     pose_ref_index: Tuple[int, int] = (0, 1),
#     crop_size: Tuple[int, int] = (300, 300),
#     use_video: bool = False,
#     video_format: str = ".mp4",
#     check_video: bool = False,
#     save_logs: bool = False,
# ) -> None:
#     """
#     Egocentric alignment of bevarioral videos.
#     Fills in the values in the "egocentric_alignment" key of the states.json file.
#     Creates training dataset for VAME at:
#     - project_name/
#         - data/
#             - filename/
#                 - filename-PE-seq.npy
#             - filename/
#                 - filename-PE-seq.npy
#     The produced .npy files contain the aligned time series data in the
#     shape of (num_dlc_features, num_video_frames).

#     Parameters
#     ---------
#     config : str
#         Path for the project config file.
#     pose_ref_index : list, optional
#         Pose reference index to be used to align. Defaults to [0, 1].
#     crop_size : tuple, optional
#         Size to crop the video. Defaults to (300,300).
#     use_video : bool, optional
#         Weather to use video to do the post alignment. Defaults to False.
#     video_format : str, optional
#         Video format, can be .mp4 or .avi. Defaults to '.mp4'.
#     check_video : bool, optional
#         Weather to check the video. Defaults to False.

#     Raises:
#     ------
#     ValueError
#         If the config.yaml indicates that the data is not egocentric.
#     """
#     try:
#         config_file = Path(config).resolve()
#         cfg = read_config(str(config_file))
#         if cfg["egocentric_data"]:
#             raise ValueError(
#                 "The config.yaml indicates that the data is egocentric. Please check the parameter 'egocentric_data'."
#             )
#         tqdm_stream = None

#         if save_logs:
#             log_path = Path(cfg["project_path"]) / "logs" / "egocentric_alignment.log"
#             logger_config.add_file_handler(str(log_path))
#             tqdm_stream = TqdmToLogger(logger=logger)

#         logger.info("Starting egocentric alignment")
#         project_path = cfg["project_path"]
#         sessions = cfg["session_names"]
#         confidence = cfg["pose_confidence"]
#         num_features = cfg["num_features"]
#         video_format = video_format
#         crop_size = crop_size

#         y_shifted_indices = np.arange(0, num_features, 2)
#         x_shifted_indices = np.arange(1, num_features, 2)
#         belly_Y_ind = pose_ref_index[0] * 2
#         belly_X_ind = (pose_ref_index[0] * 2) + 1

#         # call function and save into your VAME data folder
#         paths_to_pose_nwb_series_data = cfg["paths_to_pose_nwb_series_data"]
#         for i, session in enumerate(sessions):
#             logger.info("Aligning session %s, Pose confidence value: %.2f" % (session, confidence))
#             egocentric_time_series, frames = alignment_legacy(
#                 project_path=project_path,
#                 session=session,
#                 pose_ref_index=pose_ref_index,
#                 video_format=video_format,
#                 crop_size=crop_size,
#                 confidence=confidence,
#                 pose_estimation_filetype=cfg["pose_estimation_filetype"],
#                 path_to_pose_nwb_series_data=(
#                     paths_to_pose_nwb_series_data
#                     if not paths_to_pose_nwb_series_data
#                     else paths_to_pose_nwb_series_data[i]
#                 ),
#                 use_video=use_video,
#                 tqdm_stream=tqdm_stream,
#             )

#             # Shifiting section added 2/29/2024 PN
#             # TODO - should this be hardcoded like that?
#             egocentric_time_series_shifted = egocentric_time_series
#             belly_Y_shift = egocentric_time_series[belly_Y_ind, :]
#             belly_X_shift = egocentric_time_series[belly_X_ind, :]

#             egocentric_time_series_shifted[y_shifted_indices, :] -= belly_Y_shift
#             egocentric_time_series_shifted[x_shifted_indices, :] -= belly_X_shift

#             # Save new shifted file
#             np.save(
#                 os.path.join(
#                     project_path,
#                     "data",
#                     "processed",
#                     session,
#                     session + "-PE-seq-legacy.npy",
#                 ),
#                 egocentric_time_series_shifted,
#             )

#         logger.info("Your data is now in the right format and you can call vame.create_trainset()")
#     except Exception as e:
#         logger.exception(f"{e}")
#         raise e
#     finally:
#         logger_config.remove_file_handler()


# @save_state(model=EgocentricAlignmentFunctionSchema)
# def egocentric_alignment(
#     config: str,
#     pose_ref_1: str = "snout",
#     pose_ref_2: str = "tailbase",
#     crop_size: Tuple[int, int] = (300, 300),
#     save_logs: bool = False,
# ) -> None:
#     """
#     Egocentric alignment of bevarioral videos.
#     Fills in the values in the "egocentric_alignment" key of the states.json file.
#     Creates training dataset for VAME at:
#     - project_name/
#         - data/
#             - filename/
#                 - filename-PE-seq.npy
#             - filename/
#                 - filename-PE-seq.npy
#     The produced .npy files contain the aligned time series data in the
#     shape of (num_dlc_features, num_video_frames).

#     Parameters
#     ---------
#     config : str
#         Path for the project config file.
#     pose_ref_index : list, optional
#         Pose reference index to be used to align. Defaults to [0, 1].
#     crop_size : tuple, optional
#         Size to crop the video. Defaults to (300,300).

#     Raises:
#     ------
#     ValueError
#         If the config.yaml indicates that the data is not egocentric.
#     """
#     try:
#         config_file = Path(config).resolve()
#         cfg = read_config(str(config_file))
#         if cfg["egocentric_data"]:
#             raise ValueError(
#                 "The config.yaml indicates that the data is egocentric. Please check the parameter 'egocentric_data'."
#             )
#         tqdm_stream = None

#         if save_logs:
#             log_path = Path(cfg["project_path"]) / "logs" / "egocentric_alignment.log"
#             logger_config.add_file_handler(str(log_path))
#             tqdm_stream = TqdmToLogger(logger=logger)

#         logger.info("Starting egocentric alignment")
#         project_path = cfg["project_path"]
#         sessions = cfg["session_names"]
#         confidence = cfg["pose_confidence"]
#         num_features = cfg["num_features"]

#         y_shifted_indices = np.arange(0, num_features, 2)
#         x_shifted_indices = np.arange(1, num_features, 2)
#         # reference_Y_ind = pose_ref_index[0] * 2
#         # reference_X_ind = (pose_ref_index[0] * 2) + 1

#         # call function and save into your VAME data folder
#         for i, session in enumerate(sessions):
#             logger.info("Aligning session %s, Pose confidence value: %.2f" % (session, confidence))
#             # read out data
#             file_path = str(Path(project_path) / "data" / "raw" / f"{session}.nc")
#             _, data_mat, ds = read_pose_estimation_file(file_path=file_path)

#             # get the coordinates for alignment from data table
#             # pose_list dimensions: (num_body_parts, num_frames, 3)
#             pose_list = []
#             for i in range(int(data_mat.shape[1] / 3)):
#                 pose_list.append(data_mat[:, i * 3 : (i + 1) * 3])

#             frame_count = ds.position.time.shape[0]
#             keypoints_names = ds.keypoints.values

#             reference_X_ind = np.where(ds.keypoints.values == pose_ref_1)[0][0] * 2
#             reference_Y_ind = reference_X_ind + 1

#             pose_ref_index = (
#                 np.where(keypoints_names == pose_ref_1)[0][0],
#                 np.where(keypoints_names == pose_ref_2)[0][0],
#             )

#             egocentric_time_series = alignment(
#                 crop_size=crop_size,
#                 pose_list=pose_list,
#                 pose_ref_index=pose_ref_index,
#                 confidence=confidence,
#                 frame_count=frame_count,
#                 tqdm_stream=tqdm_stream,
#             )

#             # Shifiting section added 2/29/2024 PN
#             egocentric_time_series_shifted = egocentric_time_series
#             reference_Y_shift = egocentric_time_series[reference_Y_ind, :]
#             reference_X_shift = egocentric_time_series[reference_X_ind, :]

#             egocentric_time_series_shifted[y_shifted_indices, :] -= reference_Y_shift
#             egocentric_time_series_shifted[x_shifted_indices, :] -= reference_X_shift

#             # Save new shifted file
#             np.save(
#                 os.path.join(
#                     project_path,
#                     "data",
#                     "processed",
#                     session,
#                     session + "-PE-seq.npy",
#                 ),
#                 egocentric_time_series_shifted,
#             )

#             # Add new variable to the dataset
#             ds["position_aligned"] = (
#                 ("time", "individuals", "keypoints", "space"),
#                 egocentric_time_series_shifted.T.reshape(frame_count, 1, len(keypoints_names), 2),
#             )
#             # save to file
#             result_file = Path(project_path) / "data" / "processed" / session / f"{session}-aligned.nc"
#             ds.to_netcdf(result_file, engine="scipy")

#         logger.info("Your data is now in the right format and you can call vame.create_trainset()")
#     except Exception as e:
#         logger.exception(f"{e}")
#         raise e
#     finally:
#         logger_config.remove_file_handler()


# def alignment(
#     crop_size: Tuple[int, int],
#     pose_list: List[np.ndarray],
#     pose_ref_index: Tuple[int, int],
#     confidence: float,
#     frame_count: int,
#     tqdm_stream: Union[TqdmToLogger, None] = None,
# ) -> np.ndarray:
#     """
#     Egocentric alignment of pose estimation data.

#     Parameters
#     ----------
#     crop_size : Tuple[int, int]
#         Size to crop the video frames.
#     pose_list : List[np.ndarray]
#         List of pose coordinates.
#     pose_ref_index : Tuple[int, int]
#         Pose reference indices.
#     confidence : float
#         Pose confidence threshold.
#     frame_count : int
#         Number of frames to align.
#     tqdm_stream : Union[TqdmToLogger, None], optional
#         Tqdm stream to log the progress. Defaults to None.

#     Returns
#     -------
#     np.ndarray
#         Aligned time series data.
#     """
#     points = []

#     # for i in pose_list:
#     #     for j in i:
#     #         if j[2] <= confidence:
#     #             j[0], j[1] = np.nan, np.nan

#     # for i in pose_list:
#     #     i = interpol_first_rows_nans(i)

#     for idx in tqdm.tqdm(
#         range(frame_count),
#         disable=not True,
#         file=tqdm_stream,
#         desc="Align frames",
#     ):
#         # Read coordinates and add border
#         pose_list_bordered = []

#         for i in pose_list:
#             pose_list_bordered.append((int(i[idx][0] + crop_size[0]), int(i[idx][1] + crop_size[1])))

#         punkte = []
#         for i in pose_ref_index:
#             coord = [
#                 pose_list_bordered[i][0],
#                 pose_list_bordered[i][1],
#             ]
#             punkte.append(coord)

#         punkte = [punkte]
#         punkte = np.asarray(punkte)

#         # calculate minimal rectangle around snout and tail
#         rect = cv.minAreaRect(punkte)

#         # change size in rect tuple structure to be equal to crop_size
#         lst = list(rect)
#         # lst[0] = center_belly
#         lst[1] = crop_size
#         rect = tuple(lst)

#         # crop image
#         shifted_points = crop_and_flip(
#             rect=rect,
#             points=pose_list_bordered,
#             ref_index=pose_ref_index,
#         )

#         points.append(shifted_points)

#     time_series = np.zeros((len(pose_list) * 2, frame_count))
#     for i in range(frame_count):
#         idx = 0
#         for j in range(len(pose_list)):
#             time_series[idx : idx + 2, i] = points[i][j]
#             idx += 2

#     return time_series


# def crop_and_flip(
#     rect: Tuple,
#     points: List[np.ndarray],
#     ref_index: Tuple[int, int],
# ) -> List[np.ndarray]:
#     """
#     Crop and flip the image based on the given rectangle and points.

#     Parameters
#     ---------
#     rect : Tuple
#         Rectangle coordinates (center, size, theta).
#     points : List[np.ndarray]
#         List of points.
#     ref_index : Tuple[int, int]
#         Reference indices for alignment.

#     Returns
#     -------
#     Tuple[np.ndarray, List[np.ndarray]]
#         Cropped and flipped image, and shifted points.
#     """
#     # Read out rect structures and convert
#     center, size, theta = rect
#     center, size = tuple(map(int, center)), tuple(map(int, size))

#     # Get rotation matrix
#     M = cv.getRotationMatrix2D(center, theta, 1)

#     # shift DLC points
#     x_diff = center[0] - size[0] // 2
#     y_diff = center[1] - size[1] // 2
#     dlc_points_shifted = []
#     for i in points:
#         point = cv.transform(np.array([[[i[0], i[1]]]]), M)[0][0]
#         point[0] -= x_diff
#         point[1] -= y_diff
#         dlc_points_shifted.append(point)

#     # check if flipped correctly, otherwise flip again
#     if dlc_points_shifted[ref_index[1]][0] >= dlc_points_shifted[ref_index[0]][0]:
#         rect = (
#             (size[0] // 2, size[0] // 2),
#             size,
#             180,
#         )  # should second value be size[1]? Is this relevant to the flip? 3/5/24 KKL
#         center, size, theta = rect
#         center, size = tuple(map(int, center)), tuple(map(int, size))

#         # Get rotation matrix
#         M = cv.getRotationMatrix2D(center, theta, 1)

#         # shift DLC points
#         x_diff = center[0] - size[0] // 2
#         y_diff = center[1] - size[1] // 2

#         points = dlc_points_shifted
#         dlc_points_shifted = []

#         for i in points:
#             point = cv.transform(np.array([[[i[0], i[1]]]]), M)[0][0]
#             point[0] -= x_diff
#             point[1] -= y_diff
#             dlc_points_shifted.append(point)

#     return dlc_points_shifted



from typing import List, Tuple
import numpy as np
import pandas as pd
import cv2 as cv
import os
import tqdm
from scipy.ndimage import median_filter

from vame.logging.logger import VameLogger


logger_config = VameLogger(__name__)
logger = logger_config.logger


def consecutive(
    data: np.ndarray,
    stepsize: int = 1,
) -> List[np.ndarray]:
    """
    Find consecutive sequences in the data array.

    Parameters
    ----------
    data : np.ndarray
        Input array.
    stepsize : int, optional
        Step size. Defaults to 1.

    Returns
    -------
    List[np.ndarray]
        List of consecutive sequences.
    """
    data = data[:]
    return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)


def nan_helper(y: np.ndarray) -> Tuple:
    """
    Identifies indices of NaN values in an array and provides a function to convert them to non-NaN indices.

    Parameters
    ----------
    y : np.ndarray
        Input array containing NaN values.

    Returns
    -------
    Tuple[np.ndarray, Union[np.ndarray, None]]
        A tuple containing two elements:
        - An array of boolean values indicating the positions of NaN values.
        - A lambda function to convert NaN indices to non-NaN indices.
    """
    return np.isnan(y), lambda z: z.nonzero()[0]


# def interpol_all_nans(arr: np.ndarray) -> np.ndarray:
#     """
#     Interpolates all NaN values in the given array.

#     Parameters
#     ---------
#     arr : np.ndarray
#         Input array containing NaN values.

#     Returns
#     -------
#     np.ndarray
#         Array with NaN values replaced by interpolated values.
#     """
#     y = np.transpose(arr)
#     nans, x = nan_helper(y)
#     y[nans] = np.interp(x(nans), x(~nans), y[~nans])
#     arr = np.transpose(y)
#     return arr


def interpol_first_rows_nans(arr: np.ndarray) -> np.ndarray:
    """
    Interpolates NaN values in the given array.

    Parameters
    ----------
    arr : np.ndarray
        Input array with NaN values.

    Returns
    -------
    np.ndarray
        Array with interpolated NaN values.
    """
    y = np.transpose(arr)
    nans, x = nan_helper(y[0])
    y[0][nans] = np.interp(x(nans), x(~nans), y[0][~nans])
    nans, x = nan_helper(y[1])
    y[1][nans] = np.interp(x(nans), x(~nans), y[1][~nans])
    arr = np.transpose(y)
    return arr


def interpolate_nans_with_pandas(data: np.ndarray) -> np.ndarray:
    """
    Interpolate NaN values along the time axis of a 3D NumPy array using Pandas.

    Parameters
    ----------
    data : numpy.ndarray
        Input 3D array of shape (time, keypoints, space).

    Returns
    -------
    numpy.ndarray:
        Array with NaN values interpolated.
    """
    for kp in range(data.shape[1]):  # Loop over keypoints dimension
        for sp in range(data.shape[2]):  # Loop over space dimension (x, y)
            series = pd.Series(data[:, kp, sp])
            series_interpolated = series.interpolate(
                method="linear",
                limit_direction="both",
                axis=0,
            )
            data[:, kp, sp] = series_interpolated.values
    return data


def crop_and_flip_legacy(
    rect: Tuple,
    src: np.ndarray,
    points: List[np.ndarray],
    ref_index: Tuple[int, int],
) -> Tuple[np.ndarray, List[np.ndarray]]:
    """
    Crop and flip the image based on the given rectangle and points.

    Parameters
    ----------
    rect : Tuple
        Rectangle coordinates (center, size, theta).
    src: np.ndarray
        Source image.
    points : List[np.ndarray]
        List of points.
    ref_index : Tuple[int, int]
        Reference indices for alignment.

    Returns
    -------
    Tuple[np.ndarray, List[np.ndarray]]
        Cropped and flipped image, and shifted points.
    """
    # Read out rect structures and convert
    center, size, theta = rect
    center, size = tuple(map(int, center)), tuple(map(int, size))

    # Get rotation matrix
    M = cv.getRotationMatrix2D(center, theta, 1)

    # shift DLC points
    x_diff = center[0] - size[0] // 2
    y_diff = center[1] - size[1] // 2
    dlc_points_shifted = []
    for i in points:
        point = cv.transform(np.array([[[i[0], i[1]]]]), M)[0][0]
        point[0] -= x_diff
        point[1] -= y_diff
        dlc_points_shifted.append(point)

    # Perform rotation on src image
    dst = cv.warpAffine(src.astype("float32"), M, src.shape[:2])
    out = cv.getRectSubPix(dst, size, center)

    # check if flipped correctly, otherwise flip again
    if dlc_points_shifted[ref_index[1]][0] >= dlc_points_shifted[ref_index[0]][0]:
        rect = (
            (size[0] // 2, size[0] // 2),
            size,
            180,
        )  # should second value be size[1]? Is this relevant to the flip? 3/5/24 KKL
        center, size, theta = rect
        center, size = tuple(map(int, center)), tuple(map(int, size))

        # Get rotation matrix
        M = cv.getRotationMatrix2D(center, theta, 1)

        # shift DLC points
        x_diff = center[0] - size[0] // 2
        y_diff = center[1] - size[1] // 2

        points = dlc_points_shifted
        dlc_points_shifted = []

        for i in points:
            point = cv.transform(np.array([[[i[0], i[1]]]]), M)[0][0]
            point[0] -= x_diff
            point[1] -= y_diff
            dlc_points_shifted.append(point)

        # Perform rotation on src image
        dst = cv.warpAffine(out.astype("float32"), M, out.shape[:2])
        out = cv.getRectSubPix(dst, size, center)
    return out, dlc_points_shifted


def background(
    project_path: str,
    session: str,
    video_path: str,
    num_frames: int = 1000,
    save_background: bool = True,
) -> np.ndarray:
    """
    Compute background image from fixed camera.

    Parameters
    ----------
    project_path : str
        Path to the project directory.
    session : str
        Name of the session.
    video_path : str
        Path to the video file.
    num_frames : int, optional
        Number of frames to use for background computation. Defaults to 1000.

    Returns
    -------
    np.ndarray
        Background image.
    """
    logger.info("Computing background image ...")

    capture = cv.VideoCapture(video_path)
    if not capture.isOpened():
        raise Exception(f"Unable to open video file: {video_path}")

    frame_count = int(capture.get(cv.CAP_PROP_FRAME_COUNT))
    ret, frame = capture.read()
    height, width, _ = frame.shape
    frames = np.zeros((height, width, num_frames))

    for i in tqdm.tqdm(
        range(num_frames),
        disable=not True,
        desc="Compute background image for session %s" % session,
    ):
        rand = np.random.choice(frame_count, replace=False)
        capture.set(1, rand)
        ret, frame = capture.read()
        gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
        frames[..., i] = gray

    logger.info("Finishing up!")
    medFrame = np.median(frames, 2)
    background = median_filter(medFrame, (5, 5))

    if save_background:
        np.save(
            os.path.join(
                project_path,
                "data",
                "processed",
                session + "-background.npy",
            ),
            background,
        )

    capture.release()
    return background



from typing import List, Optional, Literal, Tuple, cast
from datetime import datetime, timezone
from pathlib import Path
import shutil
import json
import os

from vame.schemas.project import ProjectSchema
from vame.schemas.states import VAMEPipelineStatesSchema
from vame.logging.logger import VameLogger
from vame.util.auxiliary import write_config, read_config
from vame.video.video import get_video_frame_rate
from vame.io.load_poses import load_pose_estimation


logger_config = VameLogger(__name__)
logger = logger_config.logger


def init_new_project(
    project_name: str,
    poses_estimations: List[str],
    source_software: Literal["DeepLabCut", "SLEAP", "LightningPose"],
    working_directory: str = ".",
    videos: Optional[List[str]] = None,
    video_type: str = ".mp4",
    fps: int | None = None,
    copy_videos: bool = False,
    paths_to_pose_nwb_series_data: Optional[str] = None,
    config_kwargs: Optional[dict] = None,
) -> Tuple[str, dict]:
    """
    Creates a new VAME project with the given parameters.
    A VAME project is a directory with the following structure:
    - project_name/
        - data/
            - raw/
                - session1.mp4
                - session1.nc
                - session2.mp4
                - session2.nc
                - ...
            - processed/
                - session1_processed.nc
                - session2_processed.nc
                - ...
        - model/
            - pretrained_model/
        - results/
            - video1/
            - video2/
            - ...
        - states/
            - states.json
        - config.yaml

    Parameters
    ----------
    project_name : str
        Project name.
    videos : List[str]
        List of videos paths to be used in the project. E.g. ['./sample_data/Session001.mp4']
    poses_estimations : List[str]
        List of pose estimation files paths to be used in the project. E.g. ['./sample_data/pose estimation/Session001.csv']
    source_software : Literal["DeepLabCut", "SLEAP", "LightningPose"]
        Source software used for pose estimation.
    working_directory : str, optional
        Working directory. Defaults to '.'.
    video_type : str, optional
        Video extension (.mp4 or .avi). Defaults to '.mp4'.
    fps : int, optional
        Sampling rate of the videos. If not passed, it will be estimated from the video file. Defaults to None.
    copy_videos : bool, optional
        If True, the videos will be copied to the project directory. If False, symbolic links will be created instead. Defaults to False.
    paths_to_pose_nwb_series_data : Optional[str], optional
        List of paths to the pose series data in nwb files. Defaults to None.
    config_kwargs : Optional[dict], optional
        Additional configuration parameters. Defaults to None.

    Returns
    -------
    Tuple[str, dict]
        Tuple containing the path to the config file and the config data.
    """
    creation_datetime = datetime.now(timezone.utc).isoformat(timespec="seconds")
    project_path = Path(working_directory).resolve() / project_name
    if project_path.exists():
        logger.info('Project "{}" already exists!'.format(project_path))
        projconfigfile = os.path.join(str(project_path), "config.yaml")
        return projconfigfile, read_config(projconfigfile)

    data_path = project_path / "data"
    data_raw_path = data_path / "raw"
    data_processed_path = data_path / "processed"
    results_path = project_path / "results"
    model_path = project_path / "model"
    model_evaluate_path = model_path / "evaluate"
    model_pretrained_path = model_path / "pretrained_model"
    for p in [
        data_path,
        data_raw_path,
        data_processed_path,
        results_path,
        model_path,
        model_pretrained_path,
        model_evaluate_path,
    ]:
        p.mkdir(parents=True)
        logger.info('Created "{}"'.format(p))

    filetype = poses_estimations[0].split(".")[-1]
    if filetype not in ("csv", "nwb", "slp", "h5"):
        raise ValueError(f"Unsupported pose estimation file type: {filetype}. Must be one of: csv, nwb, slp, h5")
    pose_estimation_filetype = cast(Literal["csv", "nwb", "slp", "h5"], filetype)

    # Session names
    pes_paths = [Path(vp).resolve() for vp in poses_estimations]
    session_names = []
    for s in pes_paths:
        session_names.append(s.stem)

    # Creates directories under project/results/
    dirs_results = [results_path / Path(i.stem) for i in pes_paths]
    for p in dirs_results:
        p.mkdir(parents=True, exist_ok=True)

    # Copy or link videos if they are provided
    if videos:
        videos_paths = []
        for i in videos:
            # Check if it is a folder  -- WE SHOULD PROBABLY REMOVE THIS OPTION
            if os.path.isdir(i):
                vids_in_dir = [os.path.join(i, vp) for vp in os.listdir(i) if video_type in vp]
                if len(vids_in_dir) == 0:
                    logger.info(f"No videos found in {i}")
                    logger.info(f"Perhaps change the video_type, which is currently set to: {video_type}")
                else:
                    videos_paths.extend(vids_in_dir)
                    logger.info(f"{len(vids_in_dir)} videos from the directory {i} were added to the project.")
            elif os.path.isfile(i):
                videos_paths.append(i)
            else:
                logger.info(f"Invalid video path: {i}")
                raise FileNotFoundError(f"Invalid video path: {i}")

        logger.info("Copying / linking the video files... \n")
        destinations = [data_raw_path / Path(vp).name for vp in videos_paths]
        for src, dst in zip(videos_paths, destinations):
            if copy_videos:
                logger.info(f"Copying {src} to {dst}")
                shutil.copy(os.fspath(src), os.fspath(dst))
            else:
                logger.info(f"Creating symbolic link from {src} to {dst}")
                os.symlink(os.fspath(src), os.fspath(dst))

        if fps is None:
            fps = get_video_frame_rate(str(videos_paths[0]))
            logger.info(f"Estimated FPS: {fps}")
    else:
        videos_paths = [None] * len(pes_paths)
        logger.info("No videos provided.")

    # Copy pose estimation data
    logger.info("Copying pose estimation raw data...\n")
    num_features_list = []
    for pes_path, video_path in zip(poses_estimations, videos_paths):
        ds = load_pose_estimation(
            pose_estimation_file=pes_path,
            video_file=video_path,
            fps=fps,
            source_software=source_software,
        )
        output_name = data_raw_path / Path(pes_path).stem
        ds.to_netcdf(
            path=f"{output_name}.nc",
            engine="scipy",
        )
        num_features_list.append(ds.space.shape[0] * ds.keypoints.shape[0])

        output_processed_name = data_processed_path / Path(pes_path).stem
        ds.to_netcdf(
            path=f"{output_processed_name}_processed.nc",
            engine="scipy",
        )

    # Set configuration parameters
    if config_kwargs is None:
        config_kwargs = {}

    unique_num_features = list(set(num_features_list))
    if len(unique_num_features) > 1:
        raise ValueError("All pose estimation files must have the same number of features.")
    config_kwargs["num_features"] = unique_num_features[0]

    # Create config.yaml file
    new_project = ProjectSchema(
        project_name=project_name,
        creation_datetime=creation_datetime,
        project_path=str(project_path),
        session_names=session_names,
        pose_estimation_filetype=pose_estimation_filetype,
        paths_to_pose_nwb_series_data=[paths_to_pose_nwb_series_data] if paths_to_pose_nwb_series_data else None,
        **config_kwargs,
    )
    cfg_data = new_project.model_dump()
    projconfigfile = os.path.join(str(project_path), "config.yaml")
    write_config(
        config_path=projconfigfile,
        config=cfg_data,
    )

    # Create states.json file
    vame_pipeline_default_schema = VAMEPipelineStatesSchema()
    vame_pipeline_default_schema_path = Path(project_path) / "states" / "states.json"
    if not vame_pipeline_default_schema_path.parent.exists():
        vame_pipeline_default_schema_path.parent.mkdir(parents=True)
    with open(vame_pipeline_default_schema_path, "w") as f:
        json.dump(vame_pipeline_default_schema.model_dump(), f, indent=4)

    logger.info(f"A VAME project has been created at {project_path}")

    return projconfigfile, cfg_data



import os
import json
import yaml
import ruamel.yaml
from pathlib import Path
from typing import Tuple, Any
from enum import Enum


def _convert_enums_to_values(obj: Any) -> Any:
    """
    Recursively converts enum values to their string representations.

    Parameters
    ----------
    obj : Any
        The object to convert.

    Returns
    -------
    Any
        The converted object with enum values replaced by their string representations.
    """
    if isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, dict):
        return {key: _convert_enums_to_values(value) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [_convert_enums_to_values(item) for item in obj]
    return obj


def create_config_template() -> Tuple[dict, ruamel.yaml.YAML]:
    """
    Creates a template for the config.yaml file.

    Returns
    -------
    Tuple[dict, ruamel.yaml.YAML]
        A tuple containing the template dictionary and the Ruamel YAML instance.
    """
    yaml_str = """\
# Project configurations
    project_name:
    model_name:
    n_clusters:
    pose_confidence:
    \n
# Project path and videos
    project_path:
    session_names:
    \n
# Data
    all_data:
    \n
# Creation of train set:
    egocentric_data:
    robust:
    iqr_factor:
    axis:
    savgol_filter:
    savgol_length:
    savgol_order:
    test_fraction:
    \n
# RNN model general hyperparameter:
    pretrained_model:
    pretrained_weights:
    num_features:
    batch_size:
    max_epochs:
    model_snapshot:
    model_convergence:
    transition_function:
    beta:
    beta_norm:
    zdims:
    learning_rate:
    time_window:
    prediction_decoder:
    prediction_steps:
    noise:
    scheduler:
    scheduler_step_size:
    scheduler_gamma:
    #Note the optimal scheduler threshold below can vary greatly (from .1-.0001) between experiments.
    #You are encouraged to read the torch.optim.ReduceLROnPlateau docs to understand the threshold to use.
    scheduler_threshold:
    softplus:
    \n
# Segmentation:
    segmentation_algorithms:
    hmm_trained: False
    load_data:
    individual_segmentation:
    random_state_kmeans:
    n_init_kmeans:
    \n
# Video writer:
    length_of_motif_video:
    \n
# UMAP parameter:
    min_dist:
    n_neighbors:
    random_state:
    num_points:
    \n
#--------------------------------------------------------
# ONLY CHANGE ANYTHING BELOW IF YOU ARE FAMILIAR WITH RNN MODELS
# RNN encoder hyperparamter:
    hidden_size_layer_1:
    hidden_size_layer_2:
    dropout_encoder:
    \n
# RNN reconstruction hyperparameter:
    hidden_size_rec:
    dropout_rec:
    n_layers:
    \n
# RNN prediction hyperparamter:
    hidden_size_pred:
    dropout_pred:
    \n
# RNN loss hyperparameter:
    mse_reconstruction_reduction:
    mse_prediction_reduction:
    kmeans_loss:
    kmeans_lambda:
    anneal_function:
    kl_start:
    annealtime:
"""
    ruamelFile = ruamel.yaml.YAML()
    cfg_file = ruamelFile.load(yaml_str)
    return (cfg_file, ruamelFile)


def read_config(config_file: str) -> dict:
    """
    Reads structured config file defining a project.

    Parameters
    ----------
    config_file : str
        Path to the config file.

    Returns
    -------
    dict
        The contents of the config file as a dictionary.
    """
    ruamelFile = ruamel.yaml.YAML()
    path = Path(config_file)
    if os.path.exists(path):
        try:
            with open(path, "r") as f:
                cfg = ruamelFile.load(f)
                curr_dir = os.path.dirname(config_file)
                if cfg["project_path"] != curr_dir:
                    cfg["project_path"] = curr_dir
                    write_config(
                        config_path=config_file,
                        config=cfg,
                    )
        except Exception as err:
            if len(err.args) > 2:
                if err.args[2] == "could not determine a constructor for the tag '!!python/tuple'":
                    with open(path, "r") as ymlfile:
                        cfg = yaml.load(ymlfile, Loader=yaml.SafeLoader)
                        write_config(
                            config_path=config_file,
                            config=cfg,
                        )
                else:
                    raise
    else:
        raise FileNotFoundError(
            "Config file is not found. Please make sure that the file exists and/or that you passed the path of the config file correctly!"
        )
    return cfg


def write_config(
    config_path: str,
    config: dict,
) -> None:
    """
    Write structured config file.

    Parameters
    ----------
    config_path : str
        Path to the config file.
    config : dict
        Dictionary containing the config data.
    """
    with open(config_path, "w") as cf:
        cfg_file, ruamelFile = create_config_template()

        # Convert any enum values to strings before writing
        config = _convert_enums_to_values(config)

        for key in config.keys():
            cfg_file[key] = config[key]
        ruamelFile.dump(cfg_file, cf)


def read_states(config: dict) -> dict:
    """
    Reads the states.json file.

    Parameters
    ----------
    config : dict
        Dictionary containing the config data.

    Returns
    -------
    dict
        The contents of the states.json file as a dictionary.
    """
    states_path = Path(config["project_path"]) / "states" / "states.json"
    with open(states_path, "r") as f:
        states = json.load(f)
    return states



from movement.sample_data import fetch_dataset_paths
from pathlib import Path


DOWNLOAD_PATH = Path("~", ".movement", "data").expanduser().resolve()
DOWNLOAD_PATH.mkdir(parents=True, exist_ok=True)

dataset_options = {
    "DeepLabCut": "DLC_single-mouse_EPM.predictions.csv",
    "SLEAP": "SLEAP_single-mouse_EPM.predictions.slp",
}


def download_sample_data(source_software: str) -> dict:
    """
    Download sample data.

    Parameters
    ----------
    source_software : str
        Source software used for pose estimation.

    Returns
    -------
    dict
        Dictionary with the paths to the downloaded sample data.
    """
    paths_dict = fetch_dataset_paths(
        filename=dataset_options[source_software],
        with_video=True,
    )

    video_path = paths_dict.get("video")
    if video_path and video_path.stem != paths_dict["poses"].stem:
        # rename video file to match pose file
        video_path = video_path.rename(
            video_path.parent / (str(paths_dict["poses"].stem) + video_path.suffix)
        )

    paths_dict["video"] = str(video_path)
    paths_dict["poses"] = str(paths_dict["poses"])
    paths_dict["frame"] = str(paths_dict["frame"])

    return paths_dict



# from pynwb import NWBHDF5IO
# from pynwb.file import NWBFile
# from hdmf.utils import LabelledDict
# import pandas as pd


# def get_pose_data_from_nwb_file(
#     nwbfile: NWBFile,
#     path_to_pose_nwb_series_data: str,
# ) -> LabelledDict:
#     """
#     Get pose data from nwb file using a inside path to the nwb data.

#     Parameters
#     ---------
#     nwbfile : NWBFile)
#         NWB file object.
#     path_to_pose_nwb_series_data : str
#         Path to the pose data inside the nwb file.

#     Returns
#     -------
#     LabelledDict
#         Pose data.
#     """
#     if not path_to_pose_nwb_series_data:
#         raise ValueError("Path to pose nwb series data is required.")
#     pose_data = nwbfile
#     for key in path_to_pose_nwb_series_data.split("/"):
#         if isinstance(pose_data, dict):
#             pose_data = pose_data.get(key)
#             continue
#         pose_data = getattr(pose_data, key)
#     return pose_data


# def get_dataframe_from_pose_nwb_file(
#     file_path: str,
#     path_to_pose_nwb_series_data: str,
# ) -> pd.DataFrame:
#     """
#     Get pose data from nwb file and return it as a pandas DataFrame.

#     Parameters
#     ---------
#     file_path : str
#         Path to the nwb file.
#     path_to_pose_nwb_series_data : str
#         Path to the pose data inside the nwb file.

#     Returns
#     -------
#     pd.DataFrame
#         Pose data as a pandas DataFrame.
#     """
#     with NWBHDF5IO(file_path, "r") as io:
#         nwbfile = io.read()
#         pose = get_pose_data_from_nwb_file(nwbfile, path_to_pose_nwb_series_data)
#         dataframes = []
#         for label, pose_series in pose.items():
#             data = pose_series.data[:]
#             confidence = pose_series.confidence[:]
#             df = pd.DataFrame(data, columns=[f"{label}_x", f"{label}_y"])
#             df[f"likelihood_{label}"] = confidence
#             dataframes.append(df)
#         final_df = pd.concat(dataframes, axis=1)
#     return final_df



import os
from pathlib import Path
import numpy as np
import cv2 as cv
import tqdm
from typing import Union
import imageio

from vame.util.auxiliary import read_config
from vame.util.cli import get_sessions_from_user_input
from vame.schemas.states import (
    save_state,
    MotifVideosFunctionSchema,
    CommunityVideosFunctionSchema,
)
from vame.logging.logger import VameLogger, TqdmToLogger
from vame.schemas.project import SegmentationAlgorithms


logger_config = VameLogger(__name__)
logger = logger_config.logger


def create_cluster_videos(
    config: dict,
    path_to_file: str,
    session: str,
    n_clusters: int,
    video_type: str,
    flag: str,
    segmentation_algorithm: SegmentationAlgorithms,
    cohort: bool = True,
    output_video_type: str = ".mp4",
    tqdm_logger_stream: Union[TqdmToLogger, None] = None,
) -> None:
    """
    Generate cluster videos and save them to filesystem on project folder.

    Parameters
    ----------
    config : dict
        Configuration parameters.
    path_to_file : str
        Path to the file.
    session : str
        Name of the session.
    n_clusters : int
        Number of clusters.
    video_type : str
        Type of input video.
    flag : str
        Flag indicating the type of video (motif or community).
    segmentation_algorithm : SegmentationAlgorithms
        Which segmentation algorithm to use. Options are 'hmm' or 'kmeans'.
    cohort : bool, optional
        Flag indicating cohort analysis. Defaults to True.
    output_video_type : str, optional
        Type of output video. Default is '.mp4'.
    tqdm_logger_stream : TqdmToLogger, optional
        Tqdm logger stream. Default is None.

    Returns
    -------
    None
    """
    if output_video_type not in [".mp4", ".avi"]:
        raise ValueError("Output video type must be either '.avi' or '.mp4'.")

    if flag == "motif":
        logger.info("Motif videos getting created for " + session + " ...")
        labels = np.load(
            os.path.join(
                path_to_file,
                str(n_clusters) + "_" + segmentation_algorithm + "_label_" + session + ".npy",
            )
        )
    if flag == "community":
        if cohort:
            logger.info("Cohort community videos getting created for " + session + " ...")
            labels = np.load(
                os.path.join(
                    path_to_file,
                    "community",
                    "cohort_community_label_" + session + ".npy",
                )
            )
        else:
            logger.info("Community videos getting created for " + session + " ...")
            labels = np.load(
                os.path.join(
                    path_to_file,
                    "community",
                    "community_label_" + session + ".npy",
                )
            )

    video_file_path = os.path.join(
        config["project_path"],
        "data",
        "raw",
        session + video_type,
    )
    capture = cv.VideoCapture(video_file_path)
    if not capture.isOpened():
        raise ValueError(f"Video capture could not be opened. Ensure the video file is valid.\n {video_file_path}")
    width = capture.get(cv.CAP_PROP_FRAME_WIDTH)
    height = capture.get(cv.CAP_PROP_FRAME_HEIGHT)
    fps = 25  # capture.get(cv.CAP_PROP_FPS)

    cluster_start = config["time_window"] / 2
    unique_labels, count_labels = np.unique(labels, return_counts=True)

    for cluster in unique_labels:
        logger.info("Cluster: %d" % (cluster))
        cluster_lbl = np.where(labels == cluster)
        cluster_lbl = cluster_lbl[0]
        if not cluster_lbl.size:
            logger.info("Cluster is empty")
            continue

        if flag == "motif":
            output = os.path.join(
                path_to_file,
                "cluster_videos",
                session + f"-motif_%d{output_video_type}" % cluster,
            )
        if flag == "community":
            output = os.path.join(
                path_to_file,
                "community_videos",
                session + f"-community_%d{output_video_type}" % cluster,
            )

        if output_video_type == ".avi":
            codec = cv.VideoWriter_fourcc("M", "J", "P", "G")
            video_writer = cv.VideoWriter(output, codec, fps, (int(width), int(height)))
        elif output_video_type == ".mp4":
            video_writer = imageio.get_writer(
                output,
                fps=fps,
                codec="h264",
                macro_block_size=None,
            )

        if len(cluster_lbl) < config["length_of_motif_video"]:
            vid_length = len(cluster_lbl)
        else:
            vid_length = config["length_of_motif_video"]

        for num in tqdm.tqdm(range(vid_length), file=tqdm_logger_stream):
            idx = cluster_lbl[num]
            capture.set(1, idx + cluster_start)
            ret, frame = capture.read()
            if output_video_type == ".avi":
                video_writer.write(frame)
            elif output_video_type == ".mp4":
                video_writer.append_data(frame)
        if output_video_type == ".avi":
            video_writer.release()
        elif output_video_type == ".mp4":
            video_writer.close()
    capture.release()


@save_state(model=MotifVideosFunctionSchema)
def motif_videos(
    config: dict,
    segmentation_algorithm: SegmentationAlgorithms,
    video_type: str = ".mp4",
    output_video_type: str = ".mp4",
    save_logs: bool = False,
) -> None:
    """
    Generate motif videos and save them to filesystem.
    Fills in the values in the "motif_videos" key of the states.json file.
    Files are saved at:
    - project_name/
        - results/
            - session/
                - model_name/
                    - segmentation_algorithm-n_clusters/
                        - cluster_videos/
                            - session-motif_0.mp4
                            - session-motif_1.mp4
                            - ...

    Parameters
    ----------
    config : dict
        Configuration parameters.
    segmentation_algorithm : SegmentationAlgorithms
        Which segmentation algorithm to use. Options are 'hmm' or 'kmeans'.
    video_type : str, optional
        Type of video. Default is '.mp4'.
    output_video_type : str, optional
        Type of output video. Default is '.mp4'.
    save_logs : bool, optional
        Save logs to filesystem. Default is False.

    Returns
    -------
    None
    """
    try:
        tqdm_logger_stream = None
        if save_logs:
            log_path = Path(config["project_path"]) / "logs" / "motif_videos.log"
            logger_config.add_file_handler(str(log_path))
            tqdm_logger_stream = TqdmToLogger(logger=logger)
        model_name = config["model_name"]
        n_clusters = config["n_clusters"]

        logger.info(f"Creating motif videos for algorithm: {segmentation_algorithm}...")

        # Get sessions
        if config["all_data"] in ["Yes", "yes"]:
            sessions = config["session_names"]
        else:
            sessions = get_sessions_from_user_input(
                cfg=config,
                action_message="write motif videos",
            )

        logger.info("Cluster size is: %d " % n_clusters)
        for session in sessions:
            path_to_file = os.path.join(
                config["project_path"],
                "results",
                session,
                model_name,
                segmentation_algorithm + "-" + str(n_clusters),
                "",
            )
            if not os.path.exists(os.path.join(path_to_file, "cluster_videos")):
                os.mkdir(os.path.join(path_to_file, "cluster_videos"))

            create_cluster_videos(
                config=config,
                path_to_file=path_to_file,
                session=session,
                n_clusters=n_clusters,
                video_type=video_type,
                flag="motif",
                segmentation_algorithm=segmentation_algorithm,
                output_video_type=output_video_type,
                tqdm_logger_stream=tqdm_logger_stream,
            )
        logger.info("All videos have been created!")
    except Exception as e:
        logger.exception(f"Error in motif_videos: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()


@save_state(model=CommunityVideosFunctionSchema)
def community_videos(
    config: dict,
    segmentation_algorithm: SegmentationAlgorithms,
    cohort: bool = True,
    video_type: str = ".mp4",
    save_logs: bool = False,
    output_video_type: str = ".mp4",
) -> None:
    """
    Generate community videos and save them to filesystem on project community_videos folder.
    Fills in the values in the "community_videos" key of the states.json file.
    Files are saved at:

    1. If cohort is True:
    TODO: Add cohort analysis

    2. If cohort is False:
    - project_name/
        - results/
            - file_name/
                - model_name/
                    - segmentation_algorithm-n_clusters/
                        - community_videos/
                            - file_name-community_0.mp4
                            - file_name-community_1.mp4
                            - ...

    Parameters
    ----------
    config : dict
        Configuration parameters.
    segmentation_algorithm : SegmentationAlgorithms
        Which segmentation algorithm to use. Options are 'hmm' or 'kmeans'.
    cohort : bool, optional
        Flag indicating cohort analysis. Defaults to True.
    video_type : str, optional
        Type of video. Default is '.mp4'.
    save_logs : bool, optional
        Save logs to filesystem. Default is False.
    output_video_type : str, optional
        Type of output video. Default is '.mp4'.

    Returns
    -------
    None
    """
    try:
        tqdm_logger_stream = None

        if save_logs:
            log_path = Path(config["project_path"]) / "logs" / "community_videos.log"
            logger_config.add_file_handler(str(log_path))
            tqdm_logger_stream = TqdmToLogger(logger=logger)
        model_name = config["model_name"]
        n_clusters = config["n_clusters"]

        # Get sessions
        if config["all_data"] in ["Yes", "yes"]:
            sessions = config["session_names"]
        else:
            sessions = get_sessions_from_user_input(
                cfg=config,
                action_message="write community videos",
            )

        logger.info("Cluster size is: %d " % n_clusters)
        for session in sessions:
            path_to_file = os.path.join(
                config["project_path"],
                "results",
                session,
                model_name,
                segmentation_algorithm + "-" + str(n_clusters),
                "",
            )
            if not os.path.exists(os.path.join(path_to_file, "community_videos")):
                os.mkdir(os.path.join(path_to_file, "community_videos"))

            create_cluster_videos(
                config=config,
                path_to_file=path_to_file,
                session=session,
                n_clusters=n_clusters,
                video_type=video_type,
                flag="community",
                segmentation_algorithm=segmentation_algorithm,
                cohort=cohort,
                tqdm_logger_stream=tqdm_logger_stream,
                output_video_type=output_video_type,
            )

        logger.info("All videos have been created!")

    except Exception as e:
        logger.exception(f"Error in community_videos: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()



from scipy.signal import savgol_filter
import numpy as np
from pathlib import Path

from vame.logging.logger import VameLogger
from vame.io.load_poses import read_pose_estimation_file


logger_config = VameLogger(__name__)
logger = logger_config.logger


def savgol_filtering(
    config: dict,
    read_from_variable: str = "position_processed",
    save_to_variable: str = "position_processed",
) -> None:
    """
    Apply Savitzky-Golay filter to the data.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    read_from_variable : str, optional
        Variable to read from the dataset.
    save_to_variable : str, optional
        Variable to save the filtered data to.

    Returns
    -------
    None
    """
    logger.info("Applying Savitzky-Golay filter...")
    project_path = config["project_path"]
    sessions = config["session_names"]

    savgol_length = config["savgol_length"]
    savgol_order = config["savgol_order"]
    for i, session in enumerate(sessions):
        logger.info(f"Session: {session}")
        # Read raw session data
        file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        _, _, ds = read_pose_estimation_file(file_path=file_path)

        # Extract processed positions values, with shape: (time, individuals, keypoints, space)
        position = np.copy(ds[read_from_variable].values)
        filtered_position = np.copy(position)
        for individual in range(position.shape[1]):
            for keypoint in range(position.shape[2]):
                for space in range(position.shape[3]):
                    series = np.copy(position[:, individual, keypoint, space])

                    # Check if all values are zero, then skip
                    if np.all(series == 0):
                        continue

                    # Apply Savitzky-Golay filter
                    filtered_position[:, individual, keypoint, space] = savgol_filter(
                        x=series,
                        window_length=savgol_length,
                        polyorder=savgol_order,
                        axis=0,
                    )

        # Update the dataset with the filtered position values
        ds[save_to_variable] = (ds[read_from_variable].dims, filtered_position)
        ds.attrs.update({"processed_filtered": True})

        # Save the filtered dataset to file
        filtered_file_path = str(Path(project_path) / "data" / "processed" / f"{session}_processed.nc")
        ds.to_netcdf(
            path=filtered_file_path,
            engine="scipy",
        )



from typing import Literal, Optional, Tuple
from pathlib import Path
from movement.io import load_poses as mio_load_poses
import xarray as xr
import numpy as np
import pandas as pd


def load_pose_estimation(
    pose_estimation_file: Path | str,
    source_software: Literal["DeepLabCut", "SLEAP", "LightningPose"],
    video_file: Optional[Path | str] = None,
    fps: Optional[int] = None,
) -> xr.Dataset:
    """
    Load pose estimation data.

    Parameters
    ----------
    pose_estimation_file : Path or str
        Path to the pose estimation file.
    video_file : Path or str
        Path to the video file.
    fps : int
        Sampling rate of the video.
    source_software : Literal["DeepLabCut", "SLEAP", "LightningPose"]
        Source software used for pose estimation.

    Returns
    -------
    ds : xarray.Dataset
        Pose estimation dataset.
    """
    ds = mio_load_poses.from_file(
        file_path=pose_estimation_file,
        source_software=source_software,
        fps=fps,
    )
    if video_file:
        ds.attrs["video_path"] = str(video_file)
    return ds


def load_vame_dataset(ds_path: Path | str) -> xr.Dataset:
    """
    Load VAME dataset.

    Parameters
    ----------
    ds_path : Path or str
        Path to the netCDF dataset.

    Returns
    -------
    xr.Dataset
        VAME dataset
    """
    # Windows will not allow opened files to be overwritten,
    # so we need to load data into memory, close the file and move on with the operations
    with xr.open_dataset(ds_path, engine="scipy") as tmp_ds:
        ds_in_memory = tmp_ds.load()  # read entire file into memory
    return ds_in_memory


def nc_to_dataframe(nc_data):
    keypoints = nc_data["keypoints"].values
    space = nc_data["space"].values

    # Flatten position data
    position_data = nc_data["position"].isel(individuals=0).values
    position_column_names = [f"{keypoint}_{sp}" for keypoint in keypoints for sp in space]
    position_flattened = position_data.reshape(position_data.shape[0], -1)

    # Create a DataFrame for position data
    position_df = pd.DataFrame(position_flattened, columns=position_column_names)

    # Extract and flatten confidence data
    confidence_data = nc_data["confidence"].isel(individuals=0).values
    confidence_column_names = [f"{keypoint}_confidence" for keypoint in keypoints]
    confidence_flattened = confidence_data.reshape(confidence_data.shape[0], -1)
    confidence_df = pd.DataFrame(confidence_flattened, columns=confidence_column_names)

    # Combine position and confidence data
    combined_df = pd.concat([position_df, confidence_df], axis=1)

    # Reorder columns: keypoint_x, keypoint_y, keypoint_confidence
    reordered_columns = []
    for keypoint in keypoints:
        reordered_columns.extend([f"{keypoint}_x", f"{keypoint}_y", f"{keypoint}_confidence"])

    combined_df = combined_df[reordered_columns]

    return combined_df


def read_pose_estimation_file(
    file_path: str,
    file_type: Optional[Literal["csv", "nwb", "slp", "h5"]] = None,
    path_to_pose_nwb_series_data: Optional[str] = None,
) -> Tuple[pd.DataFrame, np.ndarray, xr.Dataset]:
    """
    Read pose estimation file.

    Parameters
    ----------
    file_path : str
        Path to the pose estimation file.
    file_type : PoseEstimationFiletype
        Type of the pose estimation file. Supported types are 'csv' and 'nwb'.
    path_to_pose_nwb_series_data : str, optional
        Path to the pose data inside the nwb file, by default None

    Returns
    -------
    Tuple[pd.DataFrame, np.ndarray]
        Tuple containing the pose estimation data as a pandas DataFrame and a numpy array.
    """
    ds = load_vame_dataset(ds_path=file_path)
    data = nc_to_dataframe(ds)
    data_mat = pd.DataFrame.to_numpy(data)
    return data, data_mat, ds
    # if file_type == PoseEstimationFiletype.csv:
    #     data = pd.read_csv(file_path, skiprows=2, index_col=0)
    #     if "coords" in data:
    #         data = data.drop(columns=["coords"], axis=1)
    #     data_mat = pd.DataFrame.to_numpy(data)
    #     return data, data_mat
    # elif file_type == PoseEstimationFiletype.nwb:
    #     if not path_to_pose_nwb_series_data:
    #         raise ValueError("Path to pose nwb series data is required.")
    #     data = get_dataframe_from_pose_nwb_file(
    #         file_path=file_path,
    #         path_to_pose_nwb_series_data=path_to_pose_nwb_series_data,
    #     )
    #     data_mat = pd.DataFrame.to_numpy(data)
    #     return data, data_mat
    # raise ValueError(f"Filetype {file_type} not supported")
